"""
MIT License

Copyright (c) 2017 cgalleguillosm

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from abc import abstractmethod, ABC
from re import compile

from accasim.utils.misc import CONSTANT, DEFAULT_SWF_PARSE_CONFIG, obj_assertion
from accasim.base.resource_manager_class import Resources


class WorkloadParserBase(ABC):
    """
    
    Workload Parser Abstract class 
    
    """    
    
    @abstractmethod
    def parse_line(self, line):
        """
        
        Parses a lines and retrieves the corresponding dictionary.
        
        :param line: A text line which keep all job data.
        
        :return: A dictionary of the parsed data. 
        """
        raise NotImplementedError()
    
class DefaultWorkloadParser(WorkloadParserBase):
    def __init__(self):
        WorkloadParserBase.__init__(self)
        """
            workloader_paser is a general class for parsing workload files.
            :param reg_exp: Dictionary where the name of the group is the key and the value
                its regular expresion. It must contain all the regular expresions in a sorted way,
                for appending one at the end of the previous and then recover the value(s) for each group.
            :param avoid_token: List of reg_exp to avoid reading lines. The lines that are avoided 
                won't be readed by the parser. 
        """
        self.reg_exp, self.avoid_tokens = DEFAULT_SWF_PARSE_CONFIG
        self.reg_exp_dict = {}
        self._compile_job_regexp()
        self._compile_infeasible_regexp()
        
    def feasible_line(self, line):
        """
            :param line: Line to be checked
            :return: True if it can be parse (it does not match to any avoid token), False otherwise.
        """
        for p in self._compiled_infeasible_regexp:
            if p.fullmatch(line):
                return False
        return True        
    
    def parse_line(self, line):
        """
            Parse a feasible line, returning a dict for all groupnames
            :param line: Line to be parsed
            
            :return: A dictionary of the parsed data.
            
        """
        if not self.feasible_line(line):
            return None
        _matches = self._compiled_job_regexp.match(line)
        if not _matches:
            return None
        _dict = {k:self.reg_exp_dict[k][1](v) for k, v in _matches.groupdict().items()}
        
        if _dict['requested_number_processors'] != -1:
            _dict['total_processors'] = _dict.pop('requested_number_processors')
        elif  _dict['allocated_processors'] != -1:
            _dict['total_processors'] = _dict.pop('allocated_processors')
        else:
            return None
        if _dict['requested_memory'] != -1:
            _dict['mem'] = _dict.pop('requested_memory')
        elif _dict['used_memory'] != -1:  
            _dict['mem'] = _dict.pop('used_memory')
        else:
            return None 
        return _dict

    def _compile_job_regexp(self):
        reg_exp = r''
        for (_key, _reg_exp) in self.reg_exp:
            self.reg_exp_dict[_key] = _reg_exp
            reg_exp += _reg_exp[0].format(_key)
        self._compiled_job_regexp = compile(reg_exp)
        
    def _compile_infeasible_regexp(self):
        self._compiled_infeasible_regexp = []
        for _token in self.avoid_tokens:
            self._compiled_infeasible_regexp.append(compile(_token))

class Reader(ABC):
    """
    This class is used to simulate the creation of jobs from HPC users.
    This is an abstract class. The main method is read, which must be implemented to return the set of next submission for the system.
    
    :Note:
    
        A default implementation is named as DefaultReader. This class read from a single file, and use a SWF parser to extract the jobs.
    
    """
    
    def __init__(self, _job_factory):
        """
        
        Reader class constructor.
        
        :Note:
        
            For a job generation at least the required attributes for a job event must be presented: *job_id*, *queued_time*, *duration*.
        
        :param _job_factory: Job Factory instance.
        
        
        """
        self.last_time = None
        self.job_factory = _job_factory
        self.submission_enabled = True
        self.loaded_jobs = []
    
    def next(self, current_time, time_points=2, stime_name='queued_time'):
        """
        
        Laods the data and generates the jobs that belongs to the corresponding next `time_points`.
        
        :param current_time: Current simulated point.
        :param time_points: Number of submission points to be loaded
        :param stime_name: Name of the attribute (key dictionary) of the submit/queue time
        
        :return: A tuple composed with an array of the next time points sorted chronologically and a Dictionary with an array for each time point {time_point: [job_1, ..., job_n]} 
        
        """
        #=======================================================================
        # For each time point a list of dictionary of jobs will maintained in 
        # the next_jobs variable.
        #=======================================================================
        time_samples = time_points
        next_points, next_jobs = self._reload_jobs(current_time, stime_name)
        while self.submission_enabled and time_samples >= 0:
            _dict = self._read(current_time)
            if not _dict:
                continue
            if self.last_time != _dict[stime_name]:
                self.last_time = _dict[stime_name]
                next_points.append(self.last_time)
                time_samples -= 1
                
            if self.last_time not in next_jobs:
                next_jobs[self.last_time] = []
                
            next_jobs[self.last_time].append(self.job_factory.factory(**_dict))
        if len(next_jobs) > time_points:
            self.loaded_jobs = next_jobs.pop(self.last_time, [])
            next_points.pop()                
            
        return (next_points, next_jobs)
    
    def _reload_jobs(self, current_time, stime_name):
        """
        
        Takes the already loaded jobs from a previous load process which exceeded the time steps requested.
        
        :param current_time:  Current simulated point.
        :param stime_name: Name of the attribute (key dictionary) of the submit/queue time
        
        :return: A tuple of an array of the previous time points and a dictionary with jobs for each time point.
        """
        jobs_dict = {}
        time_points = []
        for _job in self.loaded_jobs:
            _subt_job = getattr(_job, stime_name)
            assert(_subt_job >= current_time), 'Error of the loading jobs process. The submission time belongs to the past. ({} is previous to {})'.format(_subt_job, current_time)
            if _subt_job not in jobs_dict:
                jobs_dict[_subt_job] = []
                time_points.append(_subt_job)
            jobs_dict[_subt_job].append(_job)
        self.loaded_jobs = []
        return (time_points, jobs_dict)
    
    @abstractmethod
    def _read(self, current_time):
        """
        
        This method must return a dictionary with all the required keys for covering the job attributes. 
        None must be returned If the data is not valid or incomplete.   
        
        :param current_time: The current time if it's needed
        
        :return: A job's dictionary
        
        """
        raise NotImplementedError()
    
    def stop_submission(self):
        """
        
        Marks as stopped the submission process. 
        
        """
        self.submission_enabled = False
    
class DefaultReader(Reader):
    """
    
    A default implementation of the reader class. 
    This implementation works reading a workload file line by line.
    
    
    """
    def __init__(self, filepath, job_factory=None, parser=None, tweak_function=None, max_lines=None, start_time=0, equivalence={}):
        """
        Class constructor
        
        :param filepath: Filepath to the workload file.
        :param job_factory: A :class:`.job_factory` object
        :param parser: An implementation of :class:`.WorkloadParserBase` object. By default, :class:`.DefaultWorkloadParser` is used to handle SWF files. 
        :param tweak_function: Function that allows to tweak a dictionary.
        :param max_lines: Optional. Number of lines to read. None for reading the entire file.
        :param equivalence: Optional. Transforms from workload format a key:value to a new key with a new value in regards of the equivalence.
        
        """
        Reader.__init__(self, job_factory)
        self.parser = None
        self.tweak_function = None
        if parser:
            if not isinstance(parser, WorkloadParserBase):
                assert(issubclass(parser, WorkloadParserBase)), 'Only :class:`.WorkloadParserBase` class can be used as parsers'
                self.parser = parser()
            else:
                assert(isinstance(parser, WorkloadParserBase)), 'Only :class:`.WorkloadParserBase` object can be used as parsers'
                self.parser = parser                
        else:
            self.parser = DefaultWorkloadParser()
            if not tweak_function:
                _resources = self.job_factory.resource_manager.system_resources()
                self.tweak_function = DefaultTweaker(start_time, _resources, equivalence)

        if tweak_function:
            assert(isinstance(tweak_function, Tweaker)), 'The tweak_function argument must be an implementation of the :class:`.Tweaker`'
            self.tweak_function = tweak_function
        elif not self.tweak_function:
            self.tweak_function = None

        self.equivalence = equivalence
        self.start_time = start_time
        self.last_line = 0
        self.max_lines = max_lines
        self.filepath = filepath
        self.file = None
        self.EOF = True
        self.open_file()
        
    def __del__(self):
        """

        :return:
        """
        if hasattr(self, 'file') and self.file:
            self.file.close()
            self.file = None   
            self.EOF = True 
        
    def open_file(self):
        """

        :return:
        """
        if self.file is None:
            self.file = open(self.filepath)
            self.EOF = False    
        return self.file
    
    def _read_next_lines(self, n_lines=1):
        """

        :param n_lines:
        :return:
        """
        if not self.EOF:
            tmp_lines = 0
            lines = [] 
            for line in self.file:
                lines.append(line[:-1])
                self.last_line += 1
                tmp_lines += 1
                if tmp_lines == n_lines or (self.max_lines and self.max_lines == self.last_line):
                    break
            if tmp_lines < n_lines or (self.max_lines and self.max_lines == self.last_line):
                self.EOF = True
                self.stop_submission()
            if self.EOF and tmp_lines == 0:
                return None 
            return lines
        return None
    
    def _read(self, current_time=0):
        """

        :param current_time:
        :return:
        """
        line = self._read_next_lines()
        # No more lines. End of File
        if not line:
            return None
        parsed_line = self.parser.parse_line(line[0])
        if not parsed_line:
            return None
        if self.tweak_function: 
            parsed_line = self.tweak_function.tweak_function(parsed_line)
        return parsed_line

class Tweaker(ABC):
    
    def __init__(self, **kwargs):
        """

        :param kwargs:
        """
        pass
    
    @abstractmethod
    def tweak_function(self, job_dict):
        """

        :param job_dict:
        :return:
        """
        pass

class DefaultTweaker(Tweaker):

    def __init__(self, start_time, system_resources, equivalence):
        """

        :param start_time:
        :param equivalence:
        """
        obj_assertion(system_resources, Resources)
        self.start_time = start_time
        self.equivalence = equivalence if equivalence else {'processor': {'core': 1}}       
        
    def tweak_function(self, _dict):
        """
        
        As in the SWF workload logs the numbers of cores are not expressed, just the number of requested processors, we have to tweak this information
        i.e we replace the number of processors by the number of requested cores.        
        
        The equivalence from processor to core is given in the system config file. As in the example, one processor contains two cores. Then the number of cores will be
        processor \* core. Besides, memory is expressed in kB per processor. 
        
        :Example:
            
        >>> "equivalence": {
        >>>     "processor": {
        >>>         "core": 2
        >>>     }
        >>> }
        
        :param _dict: Dictionary to be tweaked.
        
        :return: The tweaked dictionary.
        
        """
        _processors = _dict['total_processors']
        
        _dict['requested_resources'] = {}
        for k, v in self.equivalence.items():
            #===================================================================
            # Since SWF doesn't provide any information about nodes request
            # every single processor request is considered as a node request.
            # An specific tweak class has to be implemented to specify a conversion of this value. 
            #===================================================================
            if k == 'processor':
                for k2, v2 in v.items():
                    _dict[k2] = v2 * _processors
                    _dict['requested_resources'][k2] = v2
        _dict['requested_nodes'] = _processors
        _dict['requested_resources']['mem'] = _dict['mem'] 
        _dict['mem'] = _dict['mem'] * _processors
        _dict['queue'] = _dict.pop('queue_number')
                
        _dict['queued_time'] = _dict.pop('queued_time') + self.start_time
        
        assert (
        _dict['core'] >= 0), 'Please consider to clean your data cannot exists requests with any info about core request.'
        assert (
        _dict['mem'] >= 0), 'Please consider to clean your data cannot exists requests with any info about mem request.'
        return _dict



"""
MIT License

Copyright (c) 2017 cgalleguillosm

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from accasim.utils.reader_class import WorkloadParserBase
from accasim.utils.file import PlainFileReader
from accasim.utils.misc import load_config, type_regexp, DEFAULT_SIMULATION
from re import compile


def define_result_parser(simulator_config=None):
    """
    This function creates a ScheduleParser instance enabled for parsing AccaSim schedule files.

    The objects produced by this function are used for post-processing simulation results.

    :return: A ScheduleParser object
    """
    try:
        if simulator_config is not None:
            _schedule_output = load_config(simulator_config)['schedule_output']
        else:
            # If no simulation config is supplied, DEFAULT_SIMULATION is used
            _schedule_output = DEFAULT_SIMULATION['SCHEDULE_OUTPUT']
        # _separators = _schedule_output['separators']
        _format = _schedule_output['format']
        _attributes = _schedule_output['attributes']
    except KeyError as e:
        print(
            'Schedule output format not identified. Please check the simulator configuration file for the key \'schedule_output\'.')
        exit()

    for _attr_name, _data_type in _attributes.items():
        _format = _format.replace('{' + _attr_name + '}', type_regexp(_data_type[-1]).format(_attr_name))
    return ScheduleParser(_format, [])


class ScheduleParser(WorkloadParserBase):
    """
    This class can read and parse schedule files produced by different sources, by personalizing the underlying parser
    object.

    ScheduleParser class is an implementation of the :class:`accasim.utils.reader_class.WorkloadParserBase`
    """

    def __init__(self, regexp, updater=[]):
        """
        Constructor for the class.

        :param regexp: Regular expression that allows to perform the parsing of lines in the schedule file
        :param updater: A list of functions used by the parser to perform automatic update operations over the lines
            that are read in the schedule file
        """
        WorkloadParserBase.__init__(self)
        self.regexp = compile(regexp)
        self.updater = updater

    def parse_line(self, line):
        """
        This method performs parsing over a single line in the schedule file, and returns a corresponding dict object.

        :param line: The line that must be parsed
        :return: A dict object is successful, None otherwise
        """
        _matches = self.regexp.match(line)
        if not _matches:
            return None
        _dict = _matches.groupdict()
        for u in self.updater:
            u(_dict)
        return _dict


class WorkloadFileReader:
    """
    Class that allows to read and parse schedule files produced by various sources.

    This class employs a ScheduleParser object in order to parse the lines that are read from the schedule file.

    """
    
    def __init__(self, workload, reg_exp, tweak_class, updater=[]):
        """
        Constructor for the class.

        :param workload: Path to the file to be read
        :param reg_exp: Regular expression used to instance the underlying ScheduleParser object
        :param tweak_class: tweak_class instance used to filter the entries read from the schedule file
        :param updater: A list of functions used by the parser to perform automatic update operations over the lines
            that are read in the schedule file
        """
        self.reader = PlainFileReader(workload)
        self.parser = ScheduleParser(reg_exp, updater)
        self.tweak = tweak_class
        
    def next(self, omit_startwith=';'):
        """
        Reads and parses one line from the workload file that is being read.

        :param omit_startwith: All lines starting with this character are skipped, and not returned
        :return: A dictionary corresponding to the read line
        """
        if self.reader.EOF:
            return None
        line = self.reader.nextline()
        if not line or line.isspace():
            return None
        try:
            while omit_startwith in line:
                line = self.reader.nextline()
                if not line:
                    return None
        except TypeError as e:
            print('Error line: {}'.format(line))
            print(e)
            exit()
        parsed_line = self.parser.parse_line(line)
        return self.tweak.tweak_function(parsed_line)



"""
MIT License

Copyright (c) 2017 cgalleguillosm

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from abc import abstractmethod, ABC

from accasim.base.event_class import EventManager
from accasim.utils.misc import CONSTANT

class AdditionalDataType:
    """
    
    Specific object type for variables added through additional data implementations 
    
    """
    def __init__(self, value):
        self.value = value

    def get(self):
        """
        :return: Returns the internal value
        
        """
        return self.value
    
    def update(self, value):
        """
        Updates the internal value
        
        :param value: New value
         
        """
        self.data = value
        
class AdditionalData(ABC):
    """

    Additional data class enables to add new behavior to the system, by executing a custom process which can use the current state of the system to create new data
    The current state of the system is maintained in the :class:`accasim.base.event_class.EventManager` object.

    """
          
    def __init__(self, event_manager=None):
        """
        
        Constructor. 
        The event mapper (:class:`accasim.base.event_class.EventManager`) must be defined at the instantiation or later, but it is mandatory for working.
        
        :param event_manager: Event manager object.
        
        """
        self.constant = CONSTANT()
        self.event_mapper = event_manager
        
    @abstractmethod
    def exec_before_dispatching(self, job_dict, queued_jobs):
        """
            Executed before the dispatcher call.
            
            :param job_dict: Dictionary containing all the loaded jobs.
            :param queued_jobs: A list containing the ids of the queued jobs.

            :return: Nothing to return. All modifications must be performed directly to the job object.
        """
        pass
    
    @abstractmethod
    def exec_after_dispatching(self, job_dict, to_dispatch_jobs, rejected_jobs):
        """
            Executed after the dispatcher call.
            
            :param job_dict: Dictionary containing all the loaded jobs.
            :param to_dispatch_jobs: A list containing the dispatching tuples [(start_time, job_id, node_allocation), ...]
            :param rejected_jobs: A list of job ids with rejected jobs by the dispatcher.
            
            :return: Nothing to return if there is no modification to the to_dispatch_jobs and rejected_jobs. 
                Otherwise these lists must be returned as tuple (to_dispatch_jobs, rejected_jobs,)  
        """
        pass
    
    @abstractmethod
    def exec_before_submission(self, **kwargs):
        pass

    @abstractmethod
    def exec_after_submission(self, **kwargs):
        pass

    @abstractmethod
    def exec_before_completion(self, **kwargs):
        pass

    @abstractmethod
    def exec_after_completion(self, removed_jobs):
        """
            Executed after releasing the resources of the jobs which have to finish at the current time. 
            
            :param removed_jobs: List of removed job object with their updated attributes.
            
            :return: Nothing 
        """
        pass

    @abstractmethod
    def stop(self):
        """
        Executed after the simulation ends.
        """
        pass
    
    def add_timepoint(self, timepoint):
        """

        Add a new time point for the events that activate the simulation steps.
        
        :param timepoint: New time point, it must be equal or later than the current time.

        """
        assert(isinstance(timepoint, int)), 'The time point must be an int'
        current_time = self.event_mapper.current_time
        if timepoint > current_time:
            tpoints = getattr(self.event_mapper, 'time_points')
            tpoints.add(timepoint)     
        
    def add_data(self, name, value):
        """

        Adds a new attribute and the respective value to the event manager.
        The new variable corresponds to a :class:`.additional_data_type`. If the variable already exists, it is updated through its update method.
        
        :param name: The name of the new attribute
        :param value: The value for the new attribute

        """
        if hasattr(self.event_mapper, name):
            var = getattr(self.event_mapper, name)
            assert(isinstance(var, AdditionalDataType)), 'Only additional_data_type class can be modified. Ensure that the {} is not already used.'.format(name)
            var.update(value)
            return
        var = AdditionalDataType(value)   
        setattr(self.event_mapper, name, var)        
        
    def set_event_manager(self, event_manager):
        """

        Set the system event manager
            
        :param event_manager: An instantiation of a :class:`accasim.base.event_class.EventManager` class or None

        """       
        if self.event_mapper:
            return
         
        self.allocator.set_resource_manager(event_manager)
        assert isinstance(event_manager, EventManager), 'Event Mapper not valid for scheduler'
        self.event_mapper = event_manager

class AdditionalDataError(Exception):
    pass



"""
MIT License

Copyright (c) 2017 cgalleguillosm, AlessioNetti

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import logging

from sys import maxsize
from random import seed
from abc import abstractmethod, ABC
from sortedcontainers.sortedlist import SortedListWithKey
from enum import Enum
from copy import deepcopy

from accasim.base.resource_manager_class import ResourceManager 
from accasim.base.allocator_class import AllocatorBase


class DispatcherError(Exception):
    pass
    
class JobVerification(Enum):
    
    REJECT = -1  # All jobs are rejected
    NO_CHECK = 0  # No verification
    CHECK_TOTAL = 1  # Total requested resources are verified  
    CHECK_REQUEST = 2  # Each node x resources are verified
  
class SchedulerBase(ABC):
    
    """
    
        This class allows to implement dispatching methods by integrating with an implementation of this class an allocator (:class:`accasim.base.allocator_class.AllocatorBase`). 
        An implementation of this class could also serve as a entire dispatching method if the allocation class is not used as default (:class:`.allocator` = None), but the resource manager must
        be set on the allocator using :func:`accasim.base.allocator_class.AllocatorBase.set_resource_manager`.
        
    """
    MAXSIZE = maxsize
    ALLOW_MAPPING_SAME_NODE = True
    
    def __init__(self, _seed, allocator=None, job_check=JobVerification.CHECK_REQUEST, **kwargs):
        """
        
        Construct a scheduler
            
        :param seed: Seed for the random state
        :param resource_manager: A Resource Manager object for dealing with system resources.
        :param allocator: Allocator object to be used by the scheduler to allocater after schedule generation. If an allocator isn't defined, the scheduler class must generate the entire dispatching plan.
        :param job_check: A job may be rejected if it doesnt comply with:
                    - JobVerification.REJECT: Any job is rejected
                    - JobVerification.NO_CHECK: All jobs are accepted
                    - JobVerification.CHECK_TOTAL: If the job requires more resources than the available in the system.
                    - JobVerification.CHECK_REQUEST: if an individual request by node requests more resources than the available one.
                    
                    
        :param kwargs:
            - skip_jobs_on_allocation: If the allocator is predefined and this parameter is true, the allocator will try to allocate jobs as much as possible. 
                Otherwise, the allocation will stop after the first fail.
                
        """
        seed(_seed)
        self._counter = 0
        self.allocator = None
        self._logger = logging.getLogger('accasim')
        self._system_capacity = None
        self._nodes_capacity = None
        self.resource_manager = None
                
        if allocator:
            assert isinstance(allocator, AllocatorBase), 'Allocator not valid for scheduler'
            self.allocator = allocator
        # self.set_resource_manager(resource_manager)

        assert(isinstance(job_check, JobVerification)), 'job_check invalid type. {}'.format(job_check.__class__)
        if job_check == JobVerification.REJECT:
            print('All jobs will be rejected, and for performance purposes the rejection messages will be omitted.')
        self._job_check = job_check
        
        # Check resources
        self._min_required_availability = kwargs.pop('min_resources', None)  # ['core', 'mem']s
        # Skip jobs during allocation
        self.skip_jobs_on_allocation = kwargs.pop('skip_jobs_on_allocation', False)
                
        
    @property
    def name(self):
        """
        
        Name of the schedulign method
        
        """
        raise NotImplementedError 
    
    @abstractmethod
    def get_id(self):
        """
        
        Must return the full ID of the scheduler, including policy and allocator.
        
        :return: the scheduler's id.
        
        """
        raise NotImplementedError
    
    @abstractmethod
    def scheduling_method(self, cur_time, es_dict, es):
        """
        
        This function must map the queued events to available nodes at the current time.
            
        :param cur_time: current time
        :param es_dict: dictionary with full data of the job events
        :param es: events to be scheduled
            
        :return a tuple of (time to schedule, event id, list of assigned nodes), an array jobs id of rejected jobs  
        
        """
        raise Exception('This function must be implemented!!')
    
    def set_resource_manager(self, resource_manager):
        """
        
        Set a resource manager. 

        :param resource_manager: An instantiation of a resource_manager class or None 
        
        """        
        if resource_manager:
            if self.allocator:
                self.allocator.set_resource_manager(resource_manager)
            assert isinstance(resource_manager, ResourceManager), 'Resource Manager not valid for scheduler'
            self.resource_manager = resource_manager
        else:
            self.resource_manager = None
            
    def schedule(self, cur_time, es_dict, es):
        """
        
        Method for schedule. It calls the specific scheduling method.
        
        :param cur_time: current time
        :param es_dict: dictionary with full data of the events
        :param es: events to be scheduled
        
        :return: a tuple of (time to schedule, event id, list of assigned nodes), array of rejected job ids.
        
        """
        assert(self.resource_manager is not None), 'The resource manager is not defined. It must defined prior to run the simulation.'

        self._counter += 1
        self._logger.debug("{} Dispatching: #{} decision".format(cur_time, self._counter))
        self._logger.debug('{} Dispatching: {} queued jobs'.format(cur_time, len(es)))
        self._logger.debug('{} Dispatching: {}'.format(cur_time, self.resource_manager.current_usage))

        rejected = []
        
        # At least a job need 1 core and 1 kb/mb/gb of mem to run
        if self._min_required_availability and any([self.resource_manager.resources.full[res] for res in self._min_required_availability]):
            self._logger.debug("There is no availability of one of the min required resource to run a job. The dispatching process will be delayed until there is enough resources.")
            return [(None, e, []) for e in es], rejected

        accepted = []
        # Verify jobs with the defined Job Policy
        for e in es:
            job = es_dict[e]
            if not job.get_checked() and not self._check_job_request(job):
                if self._job_check != JobVerification.REJECT:
                    self._logger.warning('{} has been rejected by the dispatcher. ({})'.format(e, self._job_check))
                rejected.append(e)
                continue
            accepted.append(job)
            
        to_allocate = []
        # On accepted jobs by policy, try to schedule with the scheduling policy
        if accepted:
            to_allocate, to_reject = self.scheduling_method(cur_time, accepted, es_dict)
            rejected += to_reject
            for e in to_reject:
                self._logger.warning('{} has been rejected by the dispatcher. (Scheduling policy)'.format(e))         
        
        # If there are scheduled jobs and an allocator defined, try to allocate the scheduled jobs. 
        if to_allocate and self.allocator:
            print('Skip ', self.skip_jobs_on_allocation)
            dispatching_plan = self.allocator.allocate(to_allocate, cur_time, skip=self.skip_jobs_on_allocation)
        else:
            dispatching_plan = to_allocate
            
        return dispatching_plan, rejected
    
    def _check_job_request(self, _job):
        """

        Simple method that checks if the loaded _job violates the system's resource constraints.

        :param _job: Job object

        :return: True if the _job is valid, false otherwise

        """
        _job.set_checked(True)
        if self._job_check == JobVerification.REJECT:
            return False
        
        elif self._job_check == JobVerification.NO_CHECK:
            return True
        
        elif self._job_check == JobVerification.CHECK_TOTAL:
            # We verify that the _job does not violate the system's resource constraints by comparing the total
            if not self._system_capacity:
                self._system_capacity = self.resource_manager.system_capacity('total')
            return not any([_job.requested_resources[res] * _job.requested_nodes > self._system_capacity[res] for res in _job.requested_resources.keys()])
                
        elif self._job_check == JobVerification.CHECK_REQUEST:
            if not self._nodes_capacity:
                self._nodes_capacity = self.resource_manager.system_capacity('nodes')
            # We verify the _job request can be fitted in the system        
            _requested_resources = _job.requested_resources
            _requested_nodes = _job.requested_nodes

            _fits = 0
            _diff_node = 0 
            for _node, _attrs in self._nodes_capacity.items():
                # How many time a request fits on the node
                _nfits = min([_attrs[_attr] // req for _attr, req in _requested_resources.items() if req > 0 ])
                # Update current number of times the current job fits in the nodes
                if _nfits > 0:
                    _fits += _nfits
                    _diff_node += 1
                    
                if self.ALLOW_MAPPING_SAME_NODE:
                    # Since _fits >> _diff_node this logical comparison is omitted.
                    if _fits >= _requested_nodes: 
                        return True
                else:
                    if _diff_node >= _requested_nodes:
                        return True
            
            return False
        raise DispatcherError('Invalid option.')    
    
    def __str__(self):
        return self.get_id()
    
class SimpleHeuristic(SchedulerBase):
    """
    
    Simple scheduler, sorts the event depending on the chosen policy.
    
    If a single job allocation fails, all subsequent jobs fail too.
    Sorting as name, sort funct parameters
    
    """

    def __init__(self, seed, allocator, name, sorting_parameters, **kwargs):
        SchedulerBase.__init__(self, seed, allocator, **kwargs)
        self.name = name
        self.sorting_parameters = sorting_parameters

    def get_id(self):
        """
        
        Returns the full ID of the scheduler, including policy and allocator.

        :return: the scheduler's id.
        
        """
        return '-'.join([self.__class__.__name__, self.name, self.allocator.get_id()])

    def scheduling_method(self, cur_time, jobs, es_dict):
        """
        
        This function must map the queued events to available nodes at the current time.
        
        :param cur_time: current time
        :param es_dict: dictionary with full data of the events
        :param es: events to be scheduled
        
        :return: a tuple of (time to schedule, event id, list of assigned nodes), an array jobs id of rejected jobs  
        
        """
        to_reject = []
               
        to_schedule = SortedListWithKey(jobs, **self.sorting_parameters)
        return to_schedule, to_reject

class FirstInFirstOut(SimpleHeuristic):
    """

    **FirstInFirstOut scheduling policy.** 
    
    The first come, first served (commonly called FirstInFirstOut ‒ first in, first out) 
    process scheduling algorithm is the simplest process scheduling algorithm. 
        
    """
    name = 'FIFO'
    """ Name of the Scheduler policy. """
    
    sorting_arguments = {
            'key': lambda x: x.queued_time
        }
    """ This sorting function allows to sort the jobs in relation of the scheduling policy. """

    def __init__(self, _allocator, _seed=0, **kwargs):
        """
        
        FirstInFirstOut Constructor
        
        """
        SimpleHeuristic.__init__(self, _seed, _allocator, self.name, self.sorting_arguments, **kwargs)
        
class LongestJobFirst(SimpleHeuristic):
    """
    
    **LJF scheduling policy.**
    
    Longest Job First (LJF) sorts the jobs, where the longest jobs are preferred over the shortest ones.  
        
    """
    name = 'LJF'
    """ Name of the Scheduler policy. """
    
    sorting_arguments = {
            'key': lambda x:-x.expected_duration
        }
    """ This sorting function allows to sort the jobs in relation of the scheduling policy. """

    def __init__(self, _allocator, _resource_manager=None, _seed=0, **kwargs):
        """
        
        LJF Constructor
        
        """
        SimpleHeuristic.__init__(self, _seed, _allocator, self.name, self.sorting_arguments, **kwargs)
        
class ShortestJobFirst(SimpleHeuristic):
    """
    
    **SJF scheduling policy.**
    
    Shortest Job First (SJF) sorts the jobs, where the shortest jobs are preferred over the longest ones.
    
    """
    name = 'SJF'
    """ Name of the Scheduler policy. """
    
    sorting_arguments = {
            'key': lambda x: x.expected_duration
        }
    """ This sorting function allows to sort the jobs in relation of the scheduling policy. """

    def __init__(self, _allocator, _resource_manager=None, _seed=0, **kwargs):
        """
    
        SJF Constructor
    
        """
        SimpleHeuristic.__init__(self, _seed, _allocator, self.name, self.sorting_arguments, **kwargs)

class EASYBackfilling(SchedulerBase):
    """
   
   EASY Backfilling scheduler.
   
   Whenever a job cannot be allocated, a reservation is made for it. After this, the following jobs are used to
   backfill the schedule, not allowing them to use the reserved nodes.
     
   This dispatching methods includes its own calls to the allocator over the dispatching process.
   Then it isn't use the auto allocator call, after the schedule generation.    
   
   """
   
    name = 'EBF'
    """ Name of the Scheduler policy. """
       
    def __init__(self, allocator, seed=0, **kwargs):
        """
   
       Easy BackFilling Constructor
      
       """
        SchedulerBase.__init__(self, seed, allocator=None, **kwargs)
        self._blocked_job_id = None
        self._reserved_slot = (None, [],)
        self.nonauto_allocator = allocator
        self.allocator_rm_set = False
        # self.nonauto_allocator.set_resource_manager(resource_manager)
       
    def get_id(self):
        """
   
       Returns the full ID of the scheduler, including policy and allocator.
       :return: the scheduler's id.
   
       """
        return '-'.join([self.name, self.nonauto_allocator.name])
 
    def scheduling_method(self, cur_time, queued_jobs, es_dict):
        """
        This function must map the queued events to available nodes at the current time.
       
        :param cur_time: current time
        :param queued_jobs: Jobs to be dispatched
        :param es_dict: dictionary with full data of the events
        
        
        :return: a list of tuples (time to schedule, event id, list of assigned nodes), and a list of rejected job ids  
        """
        if not self.allocator_rm_set:
            self.nonauto_allocator.set_resource_manager(self.resource_manager)
            self.allocator_rm_set = True   

                   
        avl_resources = self.resource_manager.current_availability
        self.nonauto_allocator.set_resources(avl_resources)
               
        to_dispatch = []
        to_reject = []
        _to_fill = []
        _prev_blocked = None
        _time_reached = False
        
        if self._reserved_slot[0] and self._reserved_slot[0] <= cur_time:
            _time_reached = True 
            # Tries to allocate the blocked job
            self._logger.trace('There is a blocked job {} with {}'.format(self._blocked_job_id, self._reserved_slot))
            # assert(self._blocked_job_id == queued_jobs[0].id), 'The first element is not the blocked one. ({} != {})'.format(self._blocked_job_id, queued_jobs[0].id)

            blocked_job = queued_jobs[0]
            queued_jobs = queued_jobs[1:]
                        
            allocation = self.nonauto_allocator.allocating_method(blocked_job, cur_time, skip=False)
                
            if allocation[-1]:
                self._logger.trace('{}: {} blocked job can be allocated. Unblocking'.format(cur_time, self._blocked_job_id))
                self._blocked_job_id = None
                self._reserved_slot = (None, [])
                _prev_blocked = [allocation]
                    
            else:
                # There are jobs still using the reserved nodes           
                self._logger.trace('{} job is still blocked. Reservation {}'.format(self._blocked_job_id, self._reserved_slot))
            # Add the current allocation for the (un)blocked job.
            to_dispatch += [allocation]
        
        if self._blocked_job_id is None and queued_jobs:
            # Tries to perform a FIFO allocation if there is no blocked job 
            # Returns the (partial) allocation and the idx for the blocked job, also sets the self._blocked_job_id var
            _allocated_jobs, blocked_idx = self._try_fifo_allocation(queued_jobs, cur_time)

            # There is a blocked job
            if not (blocked_idx is None):
                # If there is no a reservation, calculate it for the blocked job
                if not self._reserved_slot[0]:
                    blocked_job = queued_jobs[blocked_idx]
                    self._logger.trace('Blocked {} Job: Calculate the reservation'.format(self._blocked_job_id))
                   
                    # Current reservation (future time, reserved nodes)
                    self._reserved_slot = self._calculate_slot(cur_time, deepcopy(avl_resources), _allocated_jobs[:blocked_idx], _prev_blocked, blocked_job, es_dict)
                    self._logger.trace('Blocked {} Job: Nodes {} are reserved at {}'.format(self._blocked_job_id, self._reserved_slot[1], self._reserved_slot[0]))
                
                # Include the blocked job                
                to_dispatch += _allocated_jobs[:blocked_idx + 1]
                _to_fill = queued_jobs[blocked_idx + 1:]
            else:
                to_dispatch += _allocated_jobs                    
        else:
            if not _time_reached:
                # The blocked job
                to_dispatch += [(None, self._blocked_job_id, [])]
                # All the remaining queued jobs
                _to_fill = queued_jobs[1:]
            else:
                # The remaining queued jobs
                _to_fill = queued_jobs
        
        if _to_fill:
            self._logger.trace('Blocked job {}. {} jobs candidates to fill the gap'.format(self._blocked_job_id, len(_to_fill)))
            # Filling the gap between cur_time and res_time
            (reserved_time, reserved_nodes) = self._reserved_slot
            filling_allocation = self.nonauto_allocator.allocating_method(_to_fill, cur_time, \
                                reserved_time=reserved_time,
                                reserved_nodes=[],
                                skip=True
                            )
            # Include the remaining jobs
            to_dispatch += filling_allocation        
        return to_dispatch, to_reject
    
    def _try_fifo_allocation(self, queued_jobs, cur_time):
        """
         Allocates as many jobs as possible using the FIFO approach. As soon as one allocation fails, all subsequent jobs fail too. 
         Then, the return tuple contains info about the allocated jobs (assigned nodes and such) and also the position of the blocked job.
        
         :param queued_jobs: List of job objects
         :param cur_time: current time
         
         :return job allocation, and position of the blocked job in the list
         
        """
       
        # Try to allocate jobs as in FIFO
        _allocated_jobs = self.nonauto_allocator.allocating_method(queued_jobs, cur_time, skip=False)
        
        # Check if there is a blocked job (a job without an allocation)
        blocked_idx = None    
        for i, (_, job_id, allocated_nodes) in enumerate(_allocated_jobs):
            if not allocated_nodes:
                self._blocked_job_id = job_id   
                blocked_idx = i
                break
        return _allocated_jobs, blocked_idx

    def _calculate_slot(self, cur_time, avl_resources, decided_allocations, prev_blocked, blocked_job, es_dict):   
        """
           Computes a reservation for the blocked job, by releasing incrementally the resources used by the running
           events and recently allocated jobs. The earliest slot in which blocked_job fits is chosen.
       
        :param avl_resources: Actual available resources
        :param decided_allocations: Allocated jobs on the current iteration.
        :param prev_blocked: Allocation corresponding to the previous blocked job which has been unblocked during this iteration
        :param blocked_jobs: Event to be fitted in the time slot
        :param es_dist: Job dictionary
       
        :return: a tuple of time of the slot and nodes
        """    
        
        current_allocations = self.resource_manager.current_allocations
        # Creates a list the jobs sorted by soonest ending time first
        future_endings = SortedListWithKey(key=lambda x:x[1])
                
        # Running jobs
        for job_id, resources in current_allocations.items():
            future_endings.add((job_id, es_dict[job_id].start_time + es_dict[job_id].expected_duration, resources))
        
        # Previous blocked job has been scheduled
        if prev_blocked:
            decided_allocations += prev_blocked
        
        # Current allocated job
        for (_, job_id, nodes) in decided_allocations:
            _dec_alloc = {}
            for node in nodes:
                if not(node in _dec_alloc):
                    _dec_alloc[node] = {k:v for k, v in es_dict[job_id].requested_resources.items()}
                else:
                    for res, v in es_dict[job_id].requested_resources.items():
                        _dec_alloc[node][res] += v
            future_endings.add((job_id, cur_time + es_dict[job_id].expected_duration, _dec_alloc))

        _required_alloc = blocked_job.requested_nodes
        _requested_resources = blocked_job.requested_resources
        _partial_alloc = {}

        # Calculate the partial allocation on the current system state
        for node, resources in avl_resources.items():
            new_alloc = min([resources[req] // _requested_resources[req] for req in _requested_resources])
            if new_alloc > 0:
                _partial_alloc[node] = new_alloc            
            
        # Calculate the partial allocation on the next future endings
        for (job_id, res_time, used_nodes) in future_endings:
            for node, used_resources in used_nodes.items():
                if not(node in avl_resources):
                    avl_resources[node] = {r:0 for r in _requested_resources}
                for r, v in used_resources.items():
                    avl_resources[node][r] += v
                
                cur_alloc = _partial_alloc.get(node, 0)
                new_alloc = min([avl_resources[node][req] // _requested_resources[req] for req in _requested_resources])
                _diff = new_alloc - cur_alloc
                if _diff > 0:
                    _partial_alloc[node] = _partial_alloc.get(node, 0) + _diff                                    
            
            # At this point the blocked job can be allocated
            if sum(_partial_alloc.values()) >= _required_alloc:
                ctimes = 0
                nodes = []
                for node, times in _partial_alloc.items():
                    ctimes += times
                    nodes.append(node)
                    if ctimes >= _required_alloc:
                        break
                return (res_time, nodes,)
        raise DispatcherError('Can\'t find the slot.... no end? :(')



"""
MIT License

Copyright (c) 2017 cgalleguillosm

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from builtins import str
from os.path import join, devnull
from accasim.base.allocator_class import AllocatorBase
from accasim.base.scheduler_class import SchedulerBase 
from accasim.base.simulator_class import Simulator
from accasim.utils.file import file_exists, dir_exists, remove_dir, find_file_by
from accasim.utils.misc import obj_assertion, list_class_assertion
from accasim.utils.plot_factory import PlotFactory
from accasim.experimentation.schedule_parser import define_result_parser
import subprocess
from string import Template
from inspect import isclass
from enum import Enum
import sys
from time import sleep


_TEMPLATE = """
import sys
import os

from accasim.base.simulator_class import Simulator
$imports
        
def define_dispatcher():
    sched_class = $sched_name
    alloc_class = $alloc_name
    kwargs = $kwargs
    if alloc_class:
        return sched_class(alloc_class(), **kwargs)
    return sched_class(**kwargs)
        
if __name__ == '__main__':
    workload = '$workload'
    sys_config = '$sys_config'
    dispatcher = define_dispatcher()
    
    sim = Simulator(workload, sys_config, dispatcher, **$sim_kwargs)
    sim.start_simulation(**$run_kwargs)
"""           

class Experiment:
    
    CUSTOM_ATTRIBUTES = {'SEPARATOR': '_SEPARATOR', 'RESULTS_FOLDER': '_RESULTS_FOLDER',
                         'SCHEDULE_PREFIX': '_SCHEDULE_PREFIX', 'BENCHMARK_PREFIX': '_BENCHMARK_PREFIX'}

    SIMULATOR_ATTRIBUTES = {'scheduling_output': True, 'pprint_output': False, 'benchmark_output': True,
                            'statistics_output': True, 'job_factory': None, 'reader': None, 'additional_data': [],
                            'save_parameters': ['RESOURCE_ORDER'], 'timeout': None, 'id': None, 'skip':False, 'LOG_LEVEL': 'INFO',
                            'EXTENDED_JOB_DESCRIPTION': False}
    RUN_SIMULATOR_ATTRIBUTES = {'system_status': False}
    
    _SEPARATOR = '_'
    _RESULTS_FOLDER = 'results/{}/{}'
    _SCHEDULE_PREFIX = 'sched-'
    _BENCHMARK_PREFIX = 'bench-'
    _SIMULATOR_PARAMS_FILENAME = 'simulator_parameters.json'
    _PLOT_INPUTS = {
        PlotFactory.SCHEDULE_CLASS: _SCHEDULE_PREFIX,
        PlotFactory.BENCHMARK_CLASS: _BENCHMARK_PREFIX
    }

    def __init__(self, name, workload, sys_config, simulator_config=None, job_factory=None,
                 reader=None, **kwargs):
        """
        Experiment class constructor.

        This class allows to run automatically a set of simulations, and then generate automatically the plots. For the plot generation,
        scheduling_output and benchmark_output must be True (by default).

        :param name: Experiment name. The results of each simulator will be placed in the results/\'name\' folder.
        :param workload: Workload file to be used in the simulation. Directly readed or used as workload generation input. This depends on the reader object.
        :param sys_config: System configuration in json format.
        :param simulator_config: Configuration of the simulator parameters.
        :param job_factory: Optional. A job factory object.
        :param reader: Optional. A workload reader object for a custom reader. A SWF workload reader is defined as default.
        :param kwargs: \*\*kwargs: Optional parameters to be set in the experiment definition.
            - SEPARATOR: '_' by default. This separator is used to separate the name of the Dispatcher (schedule+allocation).
            - RESULTS_FOLDER: 'results/{}/{}' by default. Results folder where the experiments will placed. This name must contain the parent result folder, and
            two place holders, the first one belongs to the experiment name and the second one to the specific simulation.
            - SCHEDULE_PREFIX: 'sched-' by default. Corresponds the prefix name of the schedule files.
            - BENCHMAR_PREFIX: 'bench-' by default. Corresponds the prefix name of the benchmark files.
            - scheduling_output: True by default.
            - pprint_output: False by default.
            - benchmark_output: True by default.
            - statistics_output: True by default.
            - save_parameters: ['resource_order']. A list of simulator parameters to be saved as simulator_parameters.json in results/{name}/{dispatching_instance}.
            - timeout: Timeout in secs.
        """
        self.name = name
        self.dispatchers = {}
        self.results = {}
        self.sys_config = file_exists(sys_config, head_message='System configuration file: ')
        self.workload = file_exists(workload, head_message='Workload file: ')
        if simulator_config is not None:
            self.simulator_config = file_exists(simulator_config, head_message='Simulator configuration file: ')
        else:
            self.simulator_config = None

        self.parser = define_result_parser(self.simulator_config)
        self._customize(**kwargs)
        self._update_simulator_attrs(**kwargs)
        self._update_run_simulator_attrs(**kwargs)

    def add_dispatcher(self, name, dispatcher):
        """
        Add a dispatcher to the set of dispatchers.

        The dispatcher must be supplied in the form of a fully instanced SchedulerBase object, which includes an
        allocator as well.

        :param name: Dispatcher name.
        :param dispatcher: (scheduler class, allocator class, kwargs). Allocator class and kwargs can be both None

        """
        obj_assertion(name, str, 'Received {} type as dispatcher name. str type expected.',
                      [dispatcher.__class__.__name__])
        if name in self.dispatchers:
            print('Dispatcher {} already set. Skipping it'.format(name))
        self.dispatchers[name] = dispatcher

    def generate_dispatchers(self, scheduler_list, allocator_list, **kwargs):
        """
        Generate a set of dispatchers from a combination of scheduler and allocation lists.

        The input is given as a list of class names for schedulers and allocators: the method will then automatically
        generate all objects corresponding to all possible combinations of the supplied schedulers and allocators.

        :param scheduler_list: List of schedulers (:class:`accasim.base.scheduler_class.SchedulerBase`).
        :param allocator_list: List of allocators (:class:`accasim.base.allocator_class.AllocatorBase`).

        """
        list_class_assertion(scheduler_list, SchedulerBase,
                             error_msg='Scheduler objects must belong only to {} subclass',
                             msg_args=[SchedulerBase.__name__])
        list_class_assertion(allocator_list, AllocatorBase,
                             error_msg='Allocator objects must belong only to {} subclass',
                             msg_args=[AllocatorBase.__name__])
        for _alloc_class in allocator_list:
            _alloc = _alloc_class()
            for _sched_class in scheduler_list:
                _name = self._generate_name(_sched_class.name, _alloc_class.name)
                self.add_dispatcher(_name, (_sched_class, _alloc_class, kwargs,))


    def _run_simulation(self, name, dispatcher, create_script):
        """
        Runs a single simulation instance, with a specified dispatcher

        :param dispatcher: A dispatcher instantiation
        """
        if create_script:
            imports = ''
            import_vars = [(dispatcher[0].__module__, dispatcher[0].__name__,), (dispatcher[1].__module__, dispatcher[1].__name__,)] + \
                    [ (v.__module__, v.__name__,) for k, v  in dispatcher[2].items() if isclass(v)]
            for vars in import_vars:
                imports += 'from {} import {}\n'.format(*vars)
                               
            template = Template(_TEMPLATE)
            script = template.substitute(
                {
                    'workload': self.workload,
                    'sys_config': self.sys_config,
                    'imports': imports,
                    'sched_name': dispatcher[0].__name__,
                    'alloc_name': dispatcher[1].__name__,
                    'kwargs': {kw: v.value if isinstance(v, Enum) else v for kw, v in dispatcher[2].items()},
                    'sim_kwargs': self.SIMULATOR_ATTRIBUTES,
                    'run_kwargs': self.RUN_SIMULATOR_ATTRIBUTES
                }
            )
            
            fp = '{}_{}.py'.format(self.name, name)
            with open(fp, 'w+') as f:
                f.write(script)
    
            executable = sys.executable
            if not executable:
                executable = 'python3'
    
            _cmd = '{},-u,{}'.format(executable, fp)
            cwd = '.'
            FNULL = open(devnull, 'w')
            p_sim = subprocess.Popen(_cmd.split(','), cwd=cwd, stderr=subprocess.STDOUT)
            p_sim.wait()
        else:
            sched_class = dispatcher[0]
            alloc_class = dispatcher[1]
            kwargs = dispatcher[2]
            _dispatcher = sched_class(alloc_class(), **kwargs) if alloc_class else sched_class(**kwargs)
            simulator = Simulator(self.workload, self.sys_config, _dispatcher,
            simulator_config=self.simulator_config, **self.SIMULATOR_ATTRIBUTES)
            simulator.start_simulation(**self.RUN_SIMULATOR_ATTRIBUTES)
        
    def run_simulation(self, generate_plot=False, generate_scripts=False, wait=10):
        """
        Starts the simulation process. Its uses each instance of dispatching method to create the experiment.
        After that all experiments are run, the comparing plots are generated the :attr:`.generate_plot` option is set as True.

        :param generate_plot: `True` if plots must be generated, otherwise `False`. 

        """
        _total = len(self.dispatchers)
        # Run n simulations depending in the number of the defined dispatchers
        for i, (_name, _dispatcher) in enumerate(self.dispatchers.items()):
            result_folder = self.create_folders(self.name, _name)
            self.results[_name] = result_folder
            self.SIMULATOR_ATTRIBUTES['id'] = '{}_{}'.format(self.name, _name)
            self.SIMULATOR_ATTRIBUTES['RESULTS_FOLDER_NAME'] = result_folder
            print('{}/{}: Starting simulation of {}'.format(i + 1, _total, _name))
            self._run_simulation(_name, _dispatcher, generate_scripts)
            print('\n')
            sleep(wait)
        
        # Generate all plots available in the plot factory class
        if generate_plot:
            self.generate_plots(self._RESULTS_FOLDER.format(self.name, ''))
            print('Plot generation finished. ')

    def create_folders(self, name, instance_name, overwrite=True):
        """
        Create a folder if does not exists. If :attr:`.overwrite` is True, the existing folder will be deleted and then created.
        
        :param name: Corresponds to the experiment name.
        :param instance_name: Name of the current simulation. By default is the name of the dispatcher.
        :param overwrite: `True` to overwrite current results on the destiniy by deleting the entire folder.

        :return: Return the path of the new folder. 
        """
        results_folder = self._RESULTS_FOLDER.format(name, instance_name)
        if overwrite:
            remove_dir(results_folder)
        dir_exists(results_folder, create=True)
        self.SIMULATOR_ATTRIBUTES['OVERWRITE'] = overwrite
        return results_folder

    def retrieve_filepaths(self, dispatcher_results, **kwargs):
        """
        Retrieve all filepath related to the simulation results. Associates that files to the experiments.

        :param dispatcher_results: Simulation results path.
        :param kwargs: Corresponds to the type of prefix of the file. 
            Currently, there are two types of plots. Benchmark with 'bench-' and Schedule with 'sched-' file prefix.
        :return: list of experiments, list of file paths 
        """
        labels = []
        filepaths = []
        for _name, _path in dispatcher_results.items():
            labels.append(_name)
            filepaths.append(join(_path, find_file_by(_path, **kwargs)))
        return labels, filepaths

    def generate_plots(self, experiment_folder):
        """
        Generate the all plots available in the :class:`accasim.utils.PlotFactory` class.

        All plots are generated using the default parameters for each plot type. If users wish to produce plots with
        custom features and attributes, a PlotFactory object with its produce_plot method must be explicitly used,
        pointing at the result files produced by the simulations.
        
        :param experiment_folder: Path where the plots will be placed.
        """
        for plot_class, plot_types in PlotFactory.PLOT_TYPES.items():
            _prefix = self._PLOT_INPUTS[plot_class]
            resultlabel, resultpath = self.retrieve_filepaths(self.results, prefix=_prefix)
            _plot_factory = PlotFactory(plot_class, self._SIMULATOR_PARAMS_FILENAME, config=self.sys_config,
                                         workload_parser=self.parser)
            _plot_factory.set_files(resultpath, resultlabel)
            _plot_factory.pre_process()
            for plot_type in plot_types:
                output_fpath_ = join(experiment_folder, '{}.pdf'.format(plot_type))
                _plot_factory.produce_plot(type=plot_type, output=output_fpath_)
                
    def _generate_name(self, _sched_name, _alloc_name):
        """
        Generate a name of the dispatcher using the scheduler and allocator names plus the :attr:`._SEPARATOR`

        :param _sched_name: Scheduler name identifier.
        :param _alloc_name: Allocation name identifier.
        :return: Dispatching name identifier.
        """
        return '{}{}{}'.format(_sched_name, self._SEPARATOR, _alloc_name)

    def _customize(self, **kwargs):
        for k, v in kwargs.items():
            if k in self.CUSTOM_ATTRIBUTES:
                setattr(self, self.CUSTOM_ATTRIBUTES[k], v)

    def _update_simulator_attrs(self, **kwargs):
        for k, v in kwargs.items():
            if k in self.SIMULATOR_ATTRIBUTES:
                self.SIMULATOR_ATTRIBUTES[k] = v
                
    def _update_run_simulator_attrs(self, **kwargs):
        for k, v in kwargs.items():
            if k in self.RUN_SIMULATOR_ATTRIBUTES:
                self.RUN_SIMULATOR_ATTRIBUTES[k] = v



"""
MIT License

Copyright (c) 2017 cgalleguillosm

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import warnings
from multiprocessing import Pool, cpu_count
from scipy import stats as _statistical_distributions
from numpy import sum, power, roll, histogram


class DistributionFitting:
    KINT64MAX = 2 ** 63 - 1
    CONTINUOUS_DISTRIBUTIONS = {
        'alpha': 'An alpha continuous random variable.',
        'anglit': 'An anglit continuous random variable.',
        'arcsine': 'An arcsine continuous random variable.',
        'argus': 'Argus distribution',
        'beta': 'A beta continuous random variable.',
        'betaprime': 'A beta prime continuous random variable.',
        'bradford': 'A Bradford continuous random variable.',
        'burr': 'A Burr (Type III) continuous random variable.',
        'burr12': 'A Burr (Type XII) continuous random variable.',
        'cauchy': 'A Cauchy continuous random variable.',
        'chi': 'A chi continuous random variable.',
        'chi2': 'A chi-squared continuous random variable.',
        'cosine': 'A cosine continuous random variable.',
        'crystalball': 'Crystalball distribution',
        'dgamma': 'A double gamma continuous random variable.',
        'dweibull': 'A double Weibull continuous random variable.',
        'erlang': 'An Erlang continuous random variable.',
        'expon': 'An exponential continuous random variable.',
        'exponnorm': 'An exponentially modified Normal continuous random variable.',
        'exponweib': 'An exponentiated Weibull continuous random variable.',
        'exponpow': 'An exponential power continuous random variable.',
        'f': 'An F continuous random variable.',
        'fatiguelife': 'A fatigue-life (Birnbaum-Saunders) continuous random variable.',
        'fisk': 'A Fisk continuous random variable.',
        'foldcauchy': 'A folded Cauchy continuous random variable.',
        'foldnorm': 'A folded normal continuous random variable.',
        'frechet_r': 'A frechet_r continuous random variable.',
        'frechet_l': 'A frechet_l continuous random variable.',
        'genlogistic': 'A generalized logistic continuous random variable.',
        'gennorm': 'A generalized normal continuous random variable.',
        'genpareto': 'A generalized Pareto continuous random variable.',
        'genexpon': 'A generalized exponential continuous random variable.',
        'genextreme': 'A generalized extreme value continuous random variable.',
        'gausshyper': 'A Gauss hypergeometric continuous random variable.',
        'gamma': 'A gamma continuous random variable.',
        'gengamma': 'A generalized gamma continuous random variable.',
        'genhalflogistic': 'A generalized half-logistic continuous random variable.',
        'gilbrat': 'A Gilbrat continuous random variable.',
        'gompertz': 'A Gompertz (or truncated Gumbel) continuous random variable.',
        'gumbel_r': 'A right-skewed Gumbel continuous random variable.',
        'gumbel_l': 'A left-skewed Gumbel continuous random variable.',
        'halfcauchy': 'A Half-Cauchy continuous random variable.',
        'halflogistic': 'A half-logistic continuous random variable.',
        'halfnorm': 'A half-normal continuous random variable.',
        'halfgennorm': 'The upper half of a generalized normal continuous random variable.',
        'hypsecant': 'A hyperbolic secant continuous random variable.',
        'invgamma': 'An inverted gamma continuous random variable.',
        'invgauss': 'An inverse Gaussian continuous random variable.',
        'invweibull': 'An inverted Weibull continuous random variable.',
        'johnsonsb': 'A Johnson SB continuous random variable.',
        'johnsonsu': 'A Johnson SU continuous random variable.',
        'kappa4': 'Kappa 4 parameter distribution.',
        'kappa3': 'Kappa 3 parameter distribution.',
        'ksone': 'General Kolmogorov-Smirnov one-sided test.',
        'kstwobign': 'Kolmogorov-Smirnov two-sided test for large N.',
        'laplace': 'A Laplace continuous random variable.',
        'levy': 'A Levy continuous random variable.',
        'levy_l': 'A left-skewed Levy continuous random variable.',
        'levy_stable': 'A Levy-stable continuous random variable.',
        'logistic': 'A logistic (or Sech-squared) continuous random variable.',
        'loggamma': 'A log gamma continuous random variable.',
        'loglaplace': 'A log-Laplace continuous random variable.',
        'lognorm': 'A lognormal continuous random variable.',
        'lomax': 'A Lomax (Pareto of the second kind) continuous random variable.',
        'maxwell': 'A Maxwell continuous random variable.',
        'mielke': 'A Mielke\'s Beta-Kappa continuous random variable.',
        'nakagami': 'A Nakagami continuous random variable.',
        'ncx2': 'A non-central chi-squared continuous random variable.',
        'ncf': 'A non-central F distribution continuous random variable.',
        'nct': 'A non-central Student\'s T continuous random variable.',
        'norm': 'A normal continuous random variable.',
        'pareto': 'A Pareto continuous random variable.',
        'pearson3': 'A pearson type III continuous random variable.',
        'powerlaw': 'A power-function continuous random variable.',
        'powerlognorm': 'A power log-normal continuous random variable.',
        'powernorm': 'A power normal continuous random variable.',
        'rdist': 'An R-distributed continuous random variable.',
        'reciprocal': 'A reciprocal continuous random variable.',
        'rayleigh': 'A Rayleigh continuous random variable.',
        'rice': 'A Rice continuous random variable.',
        'recipinvgauss': 'A reciprocal inverse Gaussian continuous random variable.',
        'semicircular': 'A semicircular continuous random variable.',
        'skewnorm': 'A skew-normal random variable.',
        't': 'A Student\'s T continuous random variable.',
        'trapz': 'A trapezoidal continuous random variable.',
        'triang': 'A triangular continuous random variable.',
        'truncexpon': 'A truncated exponential continuous random variable.',
        'truncnorm': 'A truncated normal continuous random variable.',
        'tukeylambda': 'A Tukey-Lamdba continuous random variable.',
        'uniform': 'A uniform continuous random variable.',
        'vonmises': 'A Von Mises continuous random variable.',
        'vonmises_line': 'A Von Mises continuous random variable.',
        'wald': 'A Wald continuous random variable.',
        'weibull_min': 'Weibull minimum continuous random variable.',
        'weibull_max': 'Weibull maximum continuous random variable.',
        'wrapcauchy': 'A wrapped Cauchy continuous random variable.'
    }

    def __init__(self, dist_names=[], blacklist=[], cpus=None, default='gamma'):
        """

        :param dist_names:
        :param blacklist:
        :param cpus:
        :param default:
        """
        assert (isinstance(dist_names, list))
        assert (isinstance(blacklist, list))
        self.dist_names = [dist_name for dist_name, description in self.CONTINUOUS_DISTRIBUTIONS.items() if
                           not (dist_names in blacklist)] if not dist_names else dist_names
        self.default_distribution = default
        self._set_cpus(cpus)

    def auto_best_fit(self, data):
        """

        :param y:
        :return:
        """
        y, x = histogram(data, bins='auto', density=True)
        self.data = data
        self.y = y
        self.x = (x + roll(x, -1))[:-1] / 2.0
        try:
            results = []
            with Pool(self.cpus) as p:
                for dist_name in self.dist_names:
                    p.apply_async(self._best_fit, (dist_name,), callback=results.append)
                p.close()
                p.join()
            return (sorted(results, key=lambda r: r[1])[0], self.x, self.y,)
        except IndexError:
            return (self._best_fit(self.default_distribution), self.x, self.y,)

    def _set_cpus(self, cpus):
        """

        :param cpus:
        :return:
        """
        if cpus:
            self.cpus = cpus

        try:
            self.cpus = cpu_count() - 1
        except NotImplementedError:
            self.cpus = 1

    def _best_fit(self, dist_name):
        """

        :param dist_name:
        :return:
        """
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore')
            dist = getattr(_statistical_distributions, dist_name)
            params = dist.fit(self.data)
            arg = params[:-2]
            loc = params[-2]
            scale = params[-1]
            pdf = dist.pdf(self.x, loc=loc, scale=scale, *arg)
            sse = sum(power(self.y - pdf, 2.0))
        return (dist_name, sse, params)



"""
MIT License

Copyright (c) 2017 cgalleguillosm

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import logging

from os import path
from abc import ABC
from pydoc import locate
from builtins import str, filter
from inspect import signature
from sortedcontainers import SortedSet, SortedDict

from accasim.utils.misc import CONSTANT, FrozenDict
from accasim.base.resource_manager_class import ResourceManager
from accasim.utils.async_writer import AsyncWriter


class AttributeType:

    def __init__(self, name, type_class=None, optional=False):
        """

        Constructor for defining a new attribute type.

        :param name: Attribute name
        :param type_class: Class type of attribute (str, int, float, etc.) for casting. If value  is already casted it is not necesary.
        :param optional: False by default. If it is True, the default value will be None and it is not required to give any value to this.

        """
        assert(isinstance(name, str))
        self.name = name
        self.type = type_class
        self.optional = optional

class Event(ABC):

    def __init__(self, job_id, queued_time, duration, requested_nodes, requested_resources):
        """

        Constructor of the basic job event.

        :param job_id: Identification of the job.
        :param queued_time: Corresponding time to the submission time to the system in unix timestamp.
        :param duration: Real duration of the job in unix timestamp.
        :param requested_nodes: Number of requested nodes
        :param requested_resources: Dictionary with the requested resources for a single node.

        """
        self.constants = CONSTANT()
        self.id = str(job_id)
        self.queued_time = queued_time
        self._requested_nodes = requested_nodes        
        self._requested_resources = requested_resources
        self.start_time = None
        self.end_time = None
        self.duration = duration
        self.end_order = 0
        self._checked = False

    def subattr(self, obj, attrs):
        """

        Internal method that reads a description, and extract the value from the object itself and return it. It is used
        for genereting the output logs. (This method is candidate to be moved into utils package.)

        :param obj: Object to be analyzed
        :param attrs: Attributes to be extracted from the object

        :return: Value of the object.

        """
        if isinstance(attrs, tuple):
            values = []
            for attr in list(attrs):
                values.append(self.subattr(obj, attr))
            return values
        sp_attr = attrs.split('.')
        if len(sp_attr) > 1:
            tmp = getattr(obj, sp_attr[0])
            return self.subattr(tmp, ''.join(sp_attr[1:]))

        try:
            if isinstance(obj, dict):
                return obj.get(sp_attr[0], 'NA')
            return getattr(obj, sp_attr[0])
        except AttributeError as e:
            return 'NA'
        
    def set_checked(self, state):
        self._checked = state

    def get_checked(self):
        return self._checked

    def __str__(self):
        return 'Job_{}'.format(self.id)

    def __repr__(self):
        return 'Job_{}'.format(self.id)
    
    @property
    def requested_nodes(self):
        return self._requested_nodes

    @property
    def requested_resources(self):
        return self._requested_resources
    
    def override_requested_nodes(self, requested_nodes):
        self._requested_nodes = requested_nodes
        
    def override_requested_resources(self, type, value):
        self._requested_resources[type] = value


class JobFactory:
    def __init__(self, resource_manager=None, job_class=Event, job_attrs=[], job_mapper={}):
        """

        :param resource_manager: The resource manager of the simulator. It is required for creating the job requests.
        :param job_class: The class to be created by the Factory. By default it uses the Event class, but any subclass of it can be used (modified versions).
        :param job_attrs: The extra attributes (AttributeType class) (already job_id, queued_time and duration are mandatory) to be set in the JobEvent class
        :param job_mapper: Rename the the old key to a new key (using the value of the job_mapper dictionary)

        """
        self.resource_manager = None
        if resource_manager:
            assert(isinstance(resource_manager, ResourceManager)), 'Only subclases of :class:`.ResourceManager` are accepted.'
            self.resource_manager = resource_manager
            self.resource_manager_setup()

        assert(issubclass(job_class, Event)), 'Only subclasses of event class are accepted. Received: {} class'.format(_class.__name__)

        if job_attrs:
            assert(isinstance(job_attrs, list)), 'jobs_attrs must be a list'
            assert(all(isinstance(attr_type, AttributeType) for attr_type in job_attrs)), 'The elements of jobs_attrs must be of :class:`.AttributeType` class.'

        self.obj_type = job_class
        self.obj_parameters = list(signature(self.obj_type).parameters)
        self.attrs_names = []
        self.mandatory_attrs = {}
        self.optional_attrs = {}
        self.job_mapper = job_mapper
        self.checked = False
        self._logger = logging.getLogger('accasim')
        self.remove_req_resources = []

        for attr in job_attrs:
            _attr_name = attr.name
            assert(_attr_name not in self.attrs_names + self.obj_parameters), '{} attribute name already set. Names must be unique'.format(_attr_name)
            if attr.optional:
                # assert(_attr_name in self.obj_parameters), '{} attribute name is mandatory.'.format(_attr_name)
                self.optional_attrs[_attr_name] = attr
            else:
                self.mandatory_attrs[_attr_name] = attr
            self.attrs_names.append(_attr_name)

    def check_requested_resources(self, job_attrs):
        """
        Checks if the requested resources in the dict include all the system resources.

        :param job_attrs: Array of job attribute names
        """
        _req_resources = job_attrs['requested_resources']
        missing_res = {r for r in self.system_resources if r not in _req_resources.keys()}
        remove_req = set(_req_resources.keys()) - set(self.system_resources) 
        
        if missing_res:
            self._logger.info('Some resources has not been included in the parser, assigning 0 to the {} resources in the job request.'.format(missing_res))
            # required = {'core', 'mem'}
            inter = missing_res & required
            if inter and len(inter) != len(required):
                self._logger.error('Some mandatory attributes are missing: {}. The simulation will stop.'.format(inter))
                exit()
            self.missing_resources = missing_res
            
        if remove_req:
            self._logger.info('Job requests contains {} resources, but there are only {} resources. All the requests will be adapted.'.format(list(_req_resources.keys()), self.system_resources))
            
        self.checked = True
        return remove_req

    def resource_manager_setup(self):
        """

        The groups and system resources types are set.

        """
        self.group_resources = self.resource_manager.groups_available_resource()
        self.system_resources = self.resource_manager.system_resource_types()

    def set_resource_manager(self, resource_manager):
        if resource_manager:
            assert(isinstance(resource_manager, ResourceManager)), 'Only subclases of :class:`.resource_manager` are accepted.'
            self.resource_manager = resource_manager
            self.resource_manager_setup()


    def factory(self, **kwargs):
        """

        Creates a job instance with the dictionary received as argument. It verifies that all attributes has been included in the kwargs.

        :param \*\*kwargs: Dictionary with the job attributes.

        :return: Returns a job instance.

        """
        assert(self.resource_manager), 'Missing resource_manager attribute. It must be added via :func:`.set_resource_manager`.'
        for _old, _new in self.job_mapper.items():
            value = kwargs.pop(_old)
            kwargs[_new] = value
        _missing = list(filter(lambda x:x not in kwargs, set(self.obj_parameters + list(self.mandatory_attrs))))
        assert(not _missing), 'Missing attributes: {}'.format(', '.join(_missing))

        _obj_attr = {k:kwargs[k] for k in self.obj_parameters}

        if not self.checked:
            self.remove_req_resources = self.check_requested_resources(_obj_attr)
        
        for r in self.remove_req_resources:
            _obj_attr['requested_resources'].pop(r)        
        
        if hasattr(self, 'missing_resources'):
            for r in self.missing_resources:
                _obj_attr['requested_resources'][r] = 0

        _tmp = self.obj_type(**_obj_attr)
        setattr(_tmp, '_dict', kwargs)
        self.add_attrs(_tmp, self.mandatory_attrs, kwargs)
        self.add_attrs(_tmp, self.optional_attrs, kwargs)

        return _tmp

    def add_attrs(self, obj, reference, values):
        """

        Sets the attributes to the job object.

        :param obj: Object to be updated
        :param reference: Attribute type of reference. It contains the name, optionality and type for casting.
        :param values: Values to be added to the object

        """
        for _attr in reference:
            _type = reference[_attr].type
            _value = None
            if not reference[_attr].optional or (_attr in values and values[_attr]):
                _value = _type(values[_attr]) if _type else values[_attr]
            setattr(obj, _attr, _value)

    def add_request(self, obj):
        """

        This method sets the request of the job, it uses the resources available of the system to define it.

        :param obj: Job object

        """
        # Calculate only if it is not present
        if not hasattr(obj, 'requested_nodes'):
            _times = 0
            for _res in self.system_resources:
                _total_request = getattr(obj, _res)
                assert(_total_request >= 0), 'The request for {} is no feasible ({}). Accepted values are equal or greater than 0. Job {} must be tweaked before re-run. See the example.'.format(_res, _total_request, obj.id)
                _times = max([_times] + [round(getattr(obj, _res) / _resources[_res]) for _resources in self.group_resources.values()])
            setattr(obj, 'requested_nodes', _times)
        if not hasattr(obj, 'requested_resources'):
            _times = getattr(obj, 'requested_nodes')
            setattr(obj, 'requested_resources', {_res: getattr(obj, _res) // _times for _res in self.system_resources})            

class EventManager:

    def __init__(self, resource_manager, dispatcher, additional_data, **kwargs):
        """

        This class coordinates events submission, queueing and ending.

        :param resource_manager: Resource manager instance
        :param \*\*kwargs: nothing for the moment.

        """
        assert(isinstance(resource_manager, ResourceManager)), 'Wrong type for the resource_manager argument.'
        self.resource_manager = resource_manager
        self.dispatcher = dispatcher
        self.additional_data = additional_data
        self.constants = CONSTANT()
        # self.debug = debug
        # Stats
        self.first_time_dispatch = None
        self.last_run_time = None
        self.slowdowns = []
        self.wtimes = []

        self.current_time = None
        self.time_points = SortedSet()
        self.events = SortedDict()
        self.loaded = SortedDict()
        self.queued = []
        self.real_ending = SortedDict()
        self.running = []
        self.finished = 0
        
        self._logger = logging.getLogger('accasim')
        
        self._writers = []
        if self.constants.SCHEDULING_OUTPUT:
            _sched_fp = path.join(self.constants.RESULTS_FOLDER_PATH, self.constants.SCHED_PREFIX + self.constants.WORKLOAD_FILENAME)
            self._sched_writer = AsyncWriter(path=_sched_fp, pre_process_fun=EventManager._schd_write_preprocessor)
            self._writers.append(self._sched_writer)
        
        if self.constants.PPRINT_OUTPUT:
            _pprint_fp = path.join(self.constants.RESULTS_FOLDER_PATH, self.constants.PPRINT_PREFIX + self.constants.WORKLOAD_FILENAME)
            self._pprint_writer = AsyncWriter(path=_pprint_fp, pre_process_fun=EventManager._schd_pprint_preprocessor)
            self._writers.append(self._pprint_writer)

    def load_events(self, es):
        """

        Jobs are loaded to the system. This is the first step for a job simulation.

        :param es: List of jobs. Jobs must be subclass of event class.

        """
        for ad in self.additional_data:
            ad.exec_before_submission(es)
               
        if isinstance(es, list):
            for e in es:
                assert(isinstance(e, Event)), 'Only subclasses of event can be simulated.'
                self.load_event(e)
        else:
            assert(isinstance(es, Event)), 'Only subclasses of event can be simulated.'
            self.load_event(es)

        for ad in self.additional_data:
            ad.exec_after_submission()

    def load_event(self, e):
        """

        Internal method for job submission.

        :param e: Single job (event subclass).

        """
        assert(isinstance(e, Event)), 'Using {}, expecting a single {}'.format(e.__class__, Event.__name__)
        if not self.current_time:
            self.current_time = e.queued_time - 1
            self.time_points.add(self.current_time)
        if self.current_time == e.queued_time:
            self.submit_event(e.id)
        elif self.current_time < e.queued_time:
            if e.queued_time not in self.loaded:
                 self.loaded[e.queued_time] = []
            self.loaded[e.queued_time].append(e.id)
            self.time_points.add(e.queued_time)
        else:
            raise EventException('Time sync problem, the actual event {} was loaded after ({}) the real submit time ({}). This a programming error, must be checked.'.format(e.id, self.current_time, e.queued_time))

    def move_to_finished(self, events_dict):
        """

        There are two time points for a job could ends, the expected one and the real one.
        The job must run until the real one is reached, then if a job is waiting to finish but is less than the
        real ending time, this value must be updated with the real one.

        :param events_dict: Actual Loaded, queued and running jobs in a dictionary {id: job object}

        :return: Array of completed jobs

        """
        _es = []
        for e_id in self.real_ending.pop(self.current_time, []):
            if e_id in self.running:
                self.running.remove(e_id)
                e = events_dict[e_id]
                self.finish_event(e)
                _es.append(e_id)
        if (not self.loaded and not self.running and not self.queued) and _es:
            self.last_run_time = self.current_time
        return _es

    def finish_event(self, e):
        """

        Internal method for Job's completion. This method sets the ending time, and make some standard calculations for statistics, such as slowdown, waiting time.
        Finally it calls the methods for output.

        :param e: Job to be completed.

        """
        e.end_time = self.current_time
        e.running_time = e.end_time - e.start_time
        e.waiting_time = e.start_time - e.queued_time
        e.slowdown = float("{0:.2f}".format((e.waiting_time + e.running_time) / e.running_time)) if e.running_time != 0 else e.waiting_time if e.waiting_time != 0 else 1.0
        self.slowdowns.append(e.slowdown)
        self.wtimes.append(e.waiting_time)
        self.finished += 1
        e.end_order = self.finished
        if self.constants.SCHEDULING_OUTPUT:
            self._sched_writer.push(e)
        if self.constants.PPRINT_OUTPUT:
            self._pprint_writer.push(e)

    def dispatch_event(self, _job, _time, _time_diff, _nodes):
        """

        Internal method for Job's dispatching. This method updates the related attributes for allocation of the job.

        :param _job: Job object
        :param _time: Time of dispatching
        :param _time_diff: Time used if dispatching processing _time must be considered.
        :param _nodes: Nodes to be allocated.

        :return: True if the allocation must be performed, false otherwise. False for jobs that have duration equal to 0

        """
        _id = _job.id
        start_time = _time + _time_diff
        
        #=======================================================================
        # started, started_ended, postponed
        # state = (0, 0, 0)
        #=======================================================================
        assert(self.current_time == start_time), 'Start _time is different to the current _time'

        # Try to allocate
        done, msg = self.resource_manager.allocate_event(_job, _nodes)
        if not done:
            self._logger.error('{} Must be postponed. Reason: {}. If you see this message many times, probably you have to check your allocation heuristic.'.format(_id, msg))
            self.submit_event(_id)
            return (0, 0, 1)

        # Continue with a successfully allocated job        
        # Update job info
        _job.start_time = start_time
        _job.assigned_nodes = _nodes

        # Initialize the output writers
        if self.first_time_dispatch == None:
            self.first_time_dispatch = start_time
            for writer in self._writers:
                writer.start()
        ans = (1, 0, 0)
        real_end_time = _job.start_time + _job.duration
        if _job.duration == 0:
            self._logger.trace('{}: {} Dispatched and Finished at the same moment. Job Lenght 0'.format(self.current_time, _id))
            ans = (0, 1, 0)
        else:
            # Include real ending time int time points
            self.time_points.add(real_end_time)

        # Move to running jobs
        self.running.append(_id)

        # Relate ending time with job id
        if real_end_time not in self.real_ending:
            self.real_ending[real_end_time] = []
        self.real_ending[real_end_time].append(_id)

        return ans    

    def submit_event(self, _id):
        """

        Internal method for Job's queueing.

        """
        self.queued.append(_id)

    def next_events(self):
        """

        Return the jobs that belongs to the next time point.

        :return: Array of jobs recently submitted + queued available at current time.

        """
        if len(self.time_points) > 0:
            self.current_time = self.time_points.pop(0)
        else:
            self._logger.trace('No more time points... but there still jobs in the queue')
            self.current_time += 1
        submitted = self.loaded.pop(self.current_time, [])
        new_queue = self.queued + submitted
        self._logger.trace('{} Next events: \n-Recently submited: {}\n-Already queued: {}'.format(self.current_time, submitted, self.queued))
        self.queued.clear()

        return new_queue

    def has_events(self):
        """

        :return: True if are loaded, queued or running jobs. False otherwise.

        """
        return (self.loaded or self.queued or self.running)
    
    def call_dispatcher(self, event_dict, events):
        """
            Call the defined dispatcher. In addition, before the dispatcher call, the exec_before_dispatching method of 
            AdditionalData objects is called passing the job dictionary and current queue job ids. After the dispatcher call, 
            it calls the exec_after_dispatching and pass the job dictionary, the dispatching tuple and the rejected list.    
            
            :return: Return a tuple with a list of jobs to dispatch and to reject.  
        """
        for ad in self.additional_data:
            ad.exec_before_dispatching(event_dict, events)
                         
        to_dispatch, rejected = self.dispatcher.schedule(self.current_time, event_dict, events)
        
        for ad in self.additional_data:
            _tmp = ad.exec_after_dispatching(event_dict, to_dispatch, rejected)
            if _tmp:
                to_dispatch, rejected = _tmp
        return to_dispatch, rejected
        

    def dispatch_events(self, event_dict, to_dispatch, time_diff, omit_timediff=True):
        """

        Internal method for processing the job's dispatching. Jobs are started if start time is equals to current time.

        :param event_dict: Actual Loaded, queued and running jobs in a dictionary {id: job object}
        :param to_dispatch: A tuple which contains the (start time, job id, nodes)
        :param time_diff: Time which takes the dispatching processing time.
        :param omit_timediff: If True the time spent in generating a decision is considered as 0. False this time is considered, and all events in \
            in [current time, current time + time diff] are moved to the new current time (current time + time diff). The latter isn't implemented. 

        :return return a tuple of (#dispatched, #Dispatched + Finished (0 duration), #postponed)
        """       
        if omit_timediff:
            time_diff = 0
        
        n_disp = 0
        n_disp_finish = 0
        n_post = 0
        
        for (_time, _id, _nodes) in to_dispatch:
            assert(isinstance(_id, str)), 'Please check your return tuple in your Dispatching method. _id must be a str type. Received wrong type: {}'.format(_id.__class__)
            assert(_time is None or _time >= self.current_time), 'Receiving wrong schedules.'

            #===================================================================
            # Time must be equal or later than current time.
            #     Equals will be dispatched in the momement, instead later ones, which will be requeued with expected ending time of the job that release the resources.
            # If the expected ending is surpass, because the job takes more time to finish, the time tuple\'s element must be None. '
            #===================================================================
            if not _nodes:
                # For blocked jobs
                if _time is not None and _time != self.current_time:
                    self.time_points.add(_time)
                #==============================================================
                # Maintaining the event in the queue
                #==============================================================
                self.submit_event(_id)
                n_post += 1
                continue
            
            #=======================================================================
            # started, started_ended, postponed
            #=======================================================================
            dispatched, ended, postponed = self.dispatch_event(event_dict[_id], _time, time_diff, _nodes)
            n_disp += dispatched
            n_disp_finish += ended
            n_post += postponed   
        
        if n_disp_finish > 0:
            self.release_ended_events(event_dict)        
        return (n_disp, n_disp_finish, n_post)

    def release_ended_events(self, event_dict):
        """

        Internal method for completed jobs. Removes from the dictionary finished jobs.

        :param event_dict: Actual Loaded, queued and running jobs in a dictionary {id: job object}

        :return: return Array list of jobs objects.

        """
        for ad in self.additional_data:
            ad.exec_before_completion()
            
        _es = self.move_to_finished(event_dict)
        
        if _es:
            _removed_jobs = []
            for _e in _es:
                self.resource_manager.remove_event(_e)
                # Freeing mem (finished events)
                _removed_jobs.append(event_dict.pop(_e))
            
            for ad in self.additional_data:
                ad.exec_after_completion(_removed_jobs)
                        
        return _es

    def simulated_status(self):
        """

        Show the current state of the system in terms of loaded, queued, running and finished jobs.

        :return: String including the system info.

        """
        return ('Loaded {}, Queued {}, Running {}, and Finished {} Jobs'.format(len(self.loaded), len(self.queued), len(self.running), self.finished))

    def availability(self):
        """

        Current availability of the system.

        :return: Return the availability of the system.

        """
        return self.resource_manager.current_availability

    def usage(self):
        """

        Current usage of the system

        :return: Return the usage of the system

        """
        return self.resource_manager.current_usage

    def simulated_current_time(self):
        """

        Current time

        :return: Return the current simulated time

        """
        return self.current_time

    def stop_writers(self):
        """
        Stops the output writer threads and closes the file streams
        """
        for i in range(len(self._writers)):
            self._writers[i].stop()
            self._writers[i] = None
        
        for ad in self.additional_data:
            ad.stop()
            
    @staticmethod
    def _schd_write_preprocessor(event):
        """
        To be used as a pre-processor for AsyncWriter objects applied to event schedules.
        Pre-processes an event object and converts it to a String representation.
        It uses the format specified in the SCHEDULE_OUTPUT constant.

        :param event: The event to be written to output
        """
        constants = CONSTANT()
        _dict = constants.SCHEDULE_OUTPUT
        _attrs = {}
        for a, av in _dict['attributes'].items():
            try:
                _attrs[a] = locate(av[-1])(*event.subattr(event, av[:-1]))
            except ValueError:
                _attrs[a] = 'NA'
        output_format = _dict['format']
        return output_format.format(**_attrs) + '\n'

    @staticmethod
    def _schd_pprint_preprocessor(event):
        """
        To be used as a pre-processor for AsyncWriter objects applied to pretty-print event schedules.
        Pre-processes an event object and converts it to a String representation.
        It uses the format specified in the PPRINT_SCHEDULE_OUTPUT constant.

        :param event: The event to be written to output
        """
        constants = CONSTANT()
        _dict = constants.PPRINT_SCHEDULE_OUTPUT
        _order = _dict['order']
        _attrs = {}
        for a, av in _dict['attributes'].items():
            try:
                _attrs[a] = locate(av[-1])(*event.subattr(event, av[:-1]))
            except ValueError:
                _attrs[a] = 'NA'
        output_format = _dict['format']
        values = [_attrs[k] for k in _order]
        if event.end_order == 1:
            return (output_format.format(*_order) + '\n', output_format.format(*values) + '\n')
        else:
            return output_format.format(*values) + '\n'

    def __str__(self):
        """

        Str representation of the event job_mapper
        
        :return: Return the current system info.

        """
        return 'Loaded: {}\nQueued: {}\nRunning: {}\nReal job finish on: {},\nFinished: {}\nNext time events: {}'.format(self.loaded, self.queued, self.running, self.real_ending, self.finished, self.time_points)

class EventException(Exception):
    pass



from setuptools import setup, find_packages
from accasim import __version__
from codecs import open
from os import path
import re, mmap
import sys

here = path.abspath(path.dirname(__file__))
package_name = 'accasim'
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()
    
setup(
    name=package_name,
    version=__version__,
    description='An HPC Workload Management Simulator',
    long_description=long_description,
    url='http://accasim.readthedocs.io/en/latest/',
    author='Cristian Galleguillos',
    author_email='cristian.galleguillos.m@mail.pucv.cl',
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: MIT License',
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    keywords='research, hpc, dispathing',
    python_requires='>=3.4',

    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    install_requires=[
        'matplotlib',
        'psutil',
        'sortedcontainers',
        'numpy>= 1.6',
        'scipy>= 1.0'],
)



"""
MIT License

Copyright (c) 2017 cgalleguillosm, AlessioNetti

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""



import socket

from threading import Thread as _Thread
from os import path as _path, remove as _remove
from json import dump as _dump, dumps as _dumps, load as _load, loads as _loads
from time import perf_counter as clock
from datetime import datetime as _datetime, timedelta as _timedelta, timezone as _timezone
from re import compile as _compile
from collections import Mapping
from itertools import islice
from builtins import int, str

# ===============================================================================
# Patterns for SWF files
# ===============================================================================
_SWF_INT_PATTERN = ('\s*(?P<{}>[-+]?\d+)', int)
_SWF_FLOAT_PATTERN = ('\s*(?P<{}>[-+]?\d+\.\d+|[-+]?\d+)', float)
_SWF_AVOID_REGEXPS = [r'^;.*']
DEFAULT_SWF_PARSE_CONFIG = (
    (
        ('job_number', _SWF_INT_PATTERN),
        ('queued_time', _SWF_INT_PATTERN),
        ('wait_time', _SWF_INT_PATTERN),
        ('duration', _SWF_INT_PATTERN),
        ('allocated_processors', _SWF_INT_PATTERN),
        ('avg_cpu_time', _SWF_FLOAT_PATTERN),
        ('used_memory', _SWF_INT_PATTERN),
        ('requested_number_processors', _SWF_INT_PATTERN),
        ('requested_time', _SWF_INT_PATTERN),
        ('requested_memory', _SWF_INT_PATTERN),
        ('status', _SWF_INT_PATTERN),
        ('user_id', _SWF_INT_PATTERN),
        ('group_id', _SWF_INT_PATTERN),
        ('executable_number', _SWF_INT_PATTERN),
        ('queue_number', _SWF_INT_PATTERN),
        ('partition_number', _SWF_INT_PATTERN),
        ('preceding_job_number', _SWF_INT_PATTERN),
        ('think_time_prejob', _SWF_INT_PATTERN)
    ), _SWF_AVOID_REGEXPS)

DEFAULT_SWF_MAPPER = {
    'job_number': 'job_id',
    'requested_time': 'expected_duration',
    'executable_number': 'executable',
    'partition_number': 'partition'
}

DEFAULT_SIMULATION = {
    """
    Default and base simulation parameters. The following parameters are loaded into the :class:`.CONSTANT`.
    This constants values can be overridden by passing as kwargs in the :class:`accasim.base.simulator_class.hpc_simulator` class instantiation.
    
    :Note:
    
        * CONFIG_FOLDER_NAME: Folder where the configuration files are.
            * "CONFIG_FOLDER_NAME": "config/"
        * RESULTS_FOLDER_NAME: Folder where the configuration files will be.
            * "RESULTS_FOLDER_NAME": "results/"
        * SCHEDULE_OUTPUT: Format of the dispatching plan file.
            * "SCHEDULE_OUTPUT": 
            
            .. code:: 

                {
                    "format": "{job_id};{user};{queue_time}__{assignations}__{start_time};{end_time};{total_nodes};{total_cpu};{total_mem};{expected_duration};",
                    "attributes": {
                        "job_id": ("id", "str"),
                        "user": ("user_id", "str"),
                        "queue_time": ("queued_time", "accasim.utils.misc.str_datetime"),
                        "start_time": ("start_time", "accasim.utils.misc.str_datetime"),
                        "end_time": ("end_time", "accasim.utils.misc.str_datetime"),
                        "assignations": ("assigned_nodes", "requested_resources", "accasim.utils.misc.str_resources"),
                        "total_nodes": ("requested_nodes", "int"),
                        "total_cpu": ("core", "int"),
                        "total_mem": ("mem", "int"),
                        "expected_duration": ("expected_duration", "int")      
                    }
                }
                
        * PPRINT_SCHEDULE_OUTPUT: Format of the dispatching plan file in pretty print version. (Human readable version).
            * "PPRINT_SCHEDULE_OUTPUT":
            
            .. code:: 
            
                {
                    "format": "{:>5} {:>15} {:^19} {:^19} {:>8} {:>8} {:>8} {:>5} {:>4} {:>10} {:<20}",
                    "order": ["n", "job_id", "start_time", "end_time", "wtime", "rtime", "slowdown", "nodes", "core", "mem", "assigned_nodes"],
                    "attributes":{
                        "n": ("end_order", "int"),
                        "job_id": ("id", "str"),
                        "start_time": ("start_time", "accasim.utils.misc.str_datetime"),
                        "end_time": ("end_time", "accasim.utils.misc.str_datetime"),
                        "wtime": ("waiting_time", "int"),
                        "rtime": ("running_time", "int"),
                        "slowdown": ("slowdown", "float"),
                        "nodes": ("requested_nodes", "int"),
                        "core": ("core", "int"),
                        "mem": ("mem", "int"),
                        "assigned_nodes": ("assigned_nodes", "accasim.utils.misc.str_nodes")
                    }
                }
                
        * SCHED_PREFIX: Prefix of the dispatching plan file.
            * "SCHED_PREFIX": "sched-"
        * PPRINT_PREFIX: Prefix of the pprint file.
            * "PPRINT_PREFIX": "pprint-"
        * STATISTICS_PREFIX: Prefix of the statistic file.
            * "STATISTICS_PREFIX": "stats-"
        * BENCHMARK_PREFIX: Prefix of the benchmark file.
            * "BENCHMARK_PREFIX": "bench-"
        * SUBMISSION_ERROR_PREFIX: Prefix of the submission error file.
            * "SUBMISSION_ERROR_PREFIX": "suberror-"
        * RESOURCE_ORDER: How resource are sorted for printing purposes.
            * "RESOURCE_ORDER": ["core", "mem"]
        * WATCH_PORT: Port used for the system status daemon.
            * "WATCH_PORT": 8999
    
    """
        "CONFIG_FOLDER_NAME": "config/",
        "RESULTS_FOLDER_NAME": "results/",
        "SCHEDULE_OUTPUT": {
            "format": "{job_id};{user};{queue_time}__{assignations}__{start_time};{end_time};{total_nodes};{total_cpu};{total_mem};{expected_duration};",
            "attributes": {
                "job_id": ("id", "str"),
                "user": ("user_id", "str"),
                "queue_time": ("queued_time", "accasim.utils.misc.str_datetime"),
                "start_time": ("start_time", "accasim.utils.misc.str_datetime"),
                "end_time": ("end_time", "accasim.utils.misc.str_datetime"),
                "assignations": ("assigned_nodes", "requested_resources", "accasim.utils.misc.str_resources"),
                "total_nodes": ("requested_nodes", "int"),
                "total_cpu": ("core", "int"),
                "total_mem": ("mem", "int"),
                "expected_duration": ("expected_duration", "int")
            }
        },
        "PPRINT_SCHEDULE_OUTPUT": {
            "format": "{:>5} {:>15} {:^19} {:^19} {:>8} {:>8} {:>8} {:>5} {:>4} {:>10} {:<20}",
            "order": ["n", "job_id", "start_time", "end_time", "wtime", "rtime", "slowdown", "nodes", "core", "mem",
                      "assigned_nodes"],
            "attributes": {
                "n": ("end_order", "int"),
                "job_id": ("id", "str"),
                "start_time": ("start_time", "accasim.utils.misc.str_datetime"),
                "end_time": ("end_time", "accasim.utils.misc.str_datetime"),
                "wtime": ("waiting_time", "int"),
                "rtime": ("running_time", "int"),
                "slowdown": ("slowdown", "float"),
                "nodes": ("requested_nodes", "int"),
                "core": ("core", "int"),
                "mem": ("mem", "int"),
                "assigned_nodes": ("assigned_nodes", "accasim.utils.misc.str_nodes")
            }
        },
        "SCHED_PREFIX": "sched-",
        "PPRINT_PREFIX": "pprint-",
        "STATISTICS_PREFIX": "stats-",
        "BENCHMARK_PREFIX": "bench-",
        "SUBMISSION_ERROR_PREFIX": "suberror-",
        "RESOURCE_ORDER": ["core", "mem"],
        "WATCH_PORT": 8999
    }

def default_sorting_function(obj1, obj2, avoid_data_tokens=[';']):
    """
    
    Function for sorting the swf files in ascending order. If one of the object belongs to avoid_data_tokens, the same order is maintained by returning 1.
    
    :param obj1: Object 1
    :param obj2: Object 2
    :param avoid_data_tokens: Tokens to avoid
    
    :return: return a positive number for maintaing the order, or a negative one to change the order.
    
    """
    if obj1[0] in avoid_data_tokens or obj2[0] in avoid_data_tokens:
        return 1
    return default_sorted_attribute(obj1) - default_sorted_attribute(obj2)


def default_sorted_attribute(workload_line, attr='submit_time', converter=None):
    """
    
    :param workload_line: A line readed from the file.
    :param attr: Attribute of the line for sorting.
    :param converter: Converter function to cast the attribute.   
    
    :return: Returns the attribute of the line. Casted if it's required. 
    
    """
    value = workload_parser(workload_line, attr)[attr]
    if converter:
        return converter(value)
    return value


def workload_parser(workload_line, attrs=None, avoid_data_tokens=[';']):
    """ 
    
        Attributes of each workload line in a SWF format (separated by space):
        
        1. job_number -- a counter field, starting from 1.
        2. submit_time -- in seconds. The earliest time the log refers to is zero, and is usually the submittal time of the first job. The lines in the log are sorted by ascending submittal times. It makes sense for jobs to also be numbered in this order.
        3. wait_time -- in seconds. The difference between the job's submit time and the time at which it actually began to run. Naturally, this is only relevant to real logs, not to models.
        4. duration -- in seconds. The wall clock time the job was running (end time minus start time).
        5. allocated_processors -- an integer. In most cases this is also the number of processors the job uses; if the job does not use all of them, we typically don't know about it.
        6. avg_cpu_time -- Time Used for both user and system, in seconds. This is the average over all processors of the CPU time used, and may therefore be smaller than the wall clock runtime. If a log contains the total CPU time used by all the processors, it is divided by the number of allocated processors to derive the average.
        7. used_memory -- in kilobytes. This is again the average per processor.
        8. requested_number_processors --- Requested Number of Processors.
        9. requested_time -- This can be either runtime (measured in wallclock seconds), or average CPU time per processor (also in seconds) -- the exact meaning is determined by a header comment. In many logs this field is used for the user runtime estimate (or upper bound) used in backfilling. If a log contains a request for total CPU time, it is divided by the number of requested processors.
        10. requested_memory -- Requested memory in kilobytes per processor.
        11. status -- 1 if the job was completed, 0 if it failed, and 5 if cancelled. If information about chekcpointing or swapping is included, other values are also possible. See usage note below. This field is meaningless for models, so would be -1.
        12. user_id -- a natural number, between one and the number of different users.
        13. group_id -- a natural number, between one and the number of different groups. Some systems control resource usage by groups rather than by individual users.
        14. executable_number -- a natural number, between one and the number of different applications appearing in the workload. in some logs, this might represent a script file used to run jobs rather than the executable directly; this should be noted in a header comment.
        15. queue_number -- a natural number, between one and the number of different queues in the system. The nature of the system's queues should be explained in a header comment. This field is where batch and interactive jobs should be differentiated: we suggest the convention of denoting interactive jobs by 0.
        16. partition_number -- a natural number, between one and the number of different partitions in the systems. The nature of the system's partitions should be explained in a header comment. For example, it is possible to use partition numbers to identify which machine in a cluster was used.
        17. preceding_job_number -- this is the number of a previous job in the workload, such that the current job can only start after the termination of this preceding job. Together with the next field, this allows the workload to include feedback as described below.
        18. think_time_prejob -- this is the number of seconds that should elapse between the termination of the preceding job and the submittal of this one.
        
        :param workload_line: A Line of the workload file
        :param attrs: List of attributes to be considered. Default None, all attributes will be considered.
        :param avoid_data_tokens: List of tokens to avoid the line
        
        :return: A dictionary with all the attributes requested. If the line is returned it means that the line has the token to avoid.     
    
    """
    if workload_line[0] in avoid_data_tokens:
        return workload_line
    _common_int_pattern = ('\s*(?P<{}>[-+]?\d+)', int)
    _common_float_pattern = ('\s*(?P<{}>[-+]?\d+\.\d+|[-+]?\d+)', float)
    _dict = {
        'job_number': _common_int_pattern,
        'submit_time': _common_int_pattern,
        'wait_time': _common_int_pattern,
        'duration': _common_int_pattern,
        'allocated_processors': _common_int_pattern,
        'avg_cpu_time': _common_float_pattern,
        'used_memory': _common_int_pattern,
        'requested_number_processors': _common_int_pattern,
        'requested_time': _common_int_pattern,
        'requested_memory': _common_int_pattern,
        'status': _common_int_pattern,
        'user_id': _common_int_pattern,
        'group_id': _common_int_pattern,
        'executable_number': _common_int_pattern,
        'queue_number': _common_int_pattern,
        'partition_number': _common_int_pattern,
        'preceding_job_number': _common_int_pattern,
        'think_time_prejob': _common_int_pattern
    }
    _sequence = _dict.keys() if not attrs else ((attrs,) if isinstance(attrs, str) else attrs)
    reg_exp = r''
    for _key in _sequence:
        reg_exp += _dict[_key][0].format(_key)
    p = _compile(reg_exp)
    _matches = p.match(workload_line)
    _dict_line = _matches.groupdict()
    return {key: _dict[key][1](_dict_line[key]) for key in _sequence}


def sort_file(input_filepath, lines=None, sort_function=default_sorting_function, avoid_data_tokens=[';'],
              output_filepath=None):
    """
    
    The input file for the simulator must be sorted by submit time. It modifies the file input file, 
    or also can be saved to a new one if the output_filepath arg is defined.
      
    :param input_filepath: Input workload file
    :param lines: Number of lines to be read. It includes all lines from the begining of the file. 
    :param sort_function: (Optional) The function that sorts the file by submit time. The user is responsable to define the correct function. If a workload with SWF format is used, by default default_sorting_function (SWF workload) is used.
    :param avoid_data_tokens: (Optional) By default avoid to modify comment lines of SWF workload.      
    :param output_filepath: (Optional) The sorted data is saves into another file (this filepath). It will not content the lines that begin with tokens of the avoid_data_tokens var.
    
    :return: A list of queued time points.  

    """
    assert (callable(sort_function))
    with open(input_filepath) as f:
        sorted_file = list(f if not lines else islice(f, lines))
        sorted_file.sort(
            key=cmp_to_key(sort_function)
        )
    if output_filepath is None:
        output_filepath = input_filepath
    queued_times = sorted_list()
    with open(output_filepath, 'w') as f:
        for line in sorted_file:
            if line[0] in avoid_data_tokens:
                f.write(line)
                continue
            _line = workload_parser(line)
            if int(_line['requested_number_processors']) == -1 and int(_line['allocated_processors']) == -1 or int(
                    _line['requested_memory']) == -1 and int(_line['used_memory']) == -1:
                continue
            qtime = default_sorted_attribute(line, 'submit_time')
            queued_times.add(qtime)
            f.write(line)
    return queued_times.get_list()


def cmp_to_key(mycmp):
    """
    
    Convert a cmp= function into a key= function
    
    """

    class k(object):
        def __init__(self, obj, *args):
            self.obj = obj

        def __lt__(self, other):
            return mycmp(self.obj, other.obj) < 0

        def __gt__(self, other):
            return mycmp(self.obj, other.obj) > 0

        def __eq__(self, other):
            return mycmp(self.obj, other.obj) == 0

        def __le__(self, other):
            return mycmp(self.obj, other.obj) <= 0

        def __ge__(self, other):
            return mycmp(self.obj, other.obj) >= 0

        def __ne__(self, other):
            return mycmp(self.obj, other.obj) != 0

    return k


def from_isodatetime_2_timestamp(dtime, hours=0, minutes=0):
    """
    
    Converts a ISO datetime to Unix Timestamp
    
    :param dtime: Datetime in YYYY-MM-DD HH:MM:SS format
    
    :param hours: Hours to adjust the timezone
    :param minutes: Minutes to adjust the timezone
    
    :return: Timestamp of the dtime 
    
    """
    p = _compile(r'(\d{4})-(\d{2})-(\d{2})\s(\d{2}):(\d{2}):(\d{2})')
    m = p.search(dtime).groups()
    # year, month, day, hour, minute, second, microsecond
    t = _datetime(year=int(m[0]), month=int(m[1]), day=int(m[2]), hour=int(m[3]), minute=int(m[4]), second=int(m[5]), tzinfo=_timezone(_timedelta(hours=hours, minutes=minutes)))  # UTC
    return int(t.timestamp())


def generate_config(config_fp, **kwargs):
    """
    
    Creates a config file.
    
    :param config_fp: Filepath to the config
    :param \*\*kwargs: Source for the config data  
    
    """
    _local = {}
    for k, v in kwargs.items():
        _local[k] = v
    with open(config_fp, 'w') as c:
        _dump(_local, c, indent=2)


def hinted_tuple_hook(obj):
    """
    
    Decoder for specific object of json files, for preserving the type of the object.
    It's used with the json.load function.
    
    """
    if '__tuple__' in obj:
        return tuple(obj['items'])
    else:
        return obj


def load_config(config_fp):
    """
    
    Loads an specific config file in json format
    
    :param config_fp: Filepath of the config file.
    
    :return: Dictionary with the configuration. 
    
    """
    _dict = None
    with open(config_fp) as c:
        _dict = _load(c, object_hook=hinted_tuple_hook)
    return _dict

def clean_results(*args):
    """

    Removes the filepaths passed as argument

    :param \*args: List of filepaths 

    """
    for fp in args:
        if _path.isfile(fp) and _path.exists(fp):
            _remove(fp)

def obj_assertion(obj, class_type, error_msg=None, msg_args=None):
    """

    :param obj:
    :param class_type:
    :param error_msg:
    :param msg_args:
    :return:
    """
    if error_msg:
        assert (isinstance(obj, class_type)), error_msg.format(*msg_args)
        return
    assert (isinstance(obj, class_type))


def list_class_assertion(_list, class_type, allow_empty=False, error_msg='List class error exception.{}', msg_args=None):
    """

    :param _list:
    :param class_type:
    :param allow_empty:
    :param error_msg:
    :param msg_args:
    :return:
    """
    if not allow_empty: 
        assert (len(_list) > 0), 'Empty list not allowed.'
    try:
        if error_msg:
            assert (
                isinstance(_list, list) and all(
                    [issubclass(_class, class_type) for _class in _list])), error_msg.format(
                '' if not msg_args else msg_args)
            return
        assert (
            isinstance(_list, list) and all([issubclass(_class, class_type) for _class in _list]))
    except TypeError:
        if error_msg:
            raise Exception(error_msg.format('' if not msg_args else msg_args))

def type_regexp(_type, new_regexp={}):
    """

    :param _type:
    :param new_regexp:
    :return:
    """
    STR_TYPE = 'str'
    INT_TYPE = 'int'

    if _type == STR_TYPE:
        return '(?P<{}>[0-9a-zA-Z_\-\.@]+)'
    if _type == INT_TYPE:
        return '(?P<{}>\d+)'
    elif _type in CUSTOM_TYPES:
        return CUSTOM_TYPES[_type]
    elif _type in new_regexp:
        return new_regexp[_type]
    else:
        raise Exception('The regular expression for the {} type is not defined.')


class Singleton(object):
    """
    
    Singleton class
    
    """
    _instances = {}

    def __new__(class_, *args, **kwargs):
        if class_ not in class_._instances:
            class_._instances[class_] = super(Singleton, class_).__new__(class_, *args, **kwargs)
        return class_._instances[class_]
       

class CONSTANT(Singleton):
    """
    
    This class allows to load all config parameters into a :class:`.Singleton` Object. 
    This object will allow access to all the parameters. The parameters could be accessed as attribute name.
    
    New attrs could be passed as dict (:func:`load_constants`) or simply with (attr, value) (:func:`load_constant`)
    
    :Example:
          
        **Program**:
        
        >>> PATH = '/path/to/'
        >>> c = CONSTANT()
        >>> c.load('PATH', PATH)
        >>> print(c.PATH)
        >>> /path/to/

    :Note:
    
        It's loaded into all base class by default!
    
    """
    _constants = []

    def load_constants(self, _dict):
        """
        
        Loads an entire dictionary into the singleton.
        
        :param _dict: Dictionary with the new parameters to load. 
        
        """
        for k, v in _dict.items():
            self.load_constant(k, v)

    def load_constant(self, k, v):
        """
        
        Load an specific parameter.
        
        :param k: Name of the parameter
        :param v: Value of the parameter
        
        """
        assert (not hasattr(self, k)), '{} already exists as constant ({}={}). Choose a new name.'.format(k, k,
                                                                                                          getattr(self,
                                                                                                                  k))
        setattr(self, k, v)
        self._constants.append(k)

    def clean_constants(self):
        for _constant in self._constants:
            delattr(self, _constant)
        self._constants = []


# ===============================================================================
# Utils Types for the hinted tuple 
# ===============================================================================

class str_:
    def __init__(self, text):
        self.text = text

    def __str__(self):
        return self.text


class str_datetime:
    REGEX = '\d{4}-\d{2}-\d{2}\s\d{2}:\d{2}:\d{2}'
    REGEX_GROUP = '(?P<{}>\d{{4}}-\d{{2}}-\d{{2}}\s\d{{2}}:\d{{2}}:\d{{2}})'

    def __init__(self, epoch_time):
        self.datetime = _datetime.fromtimestamp(int(epoch_time))
        self.str_datetime = self.datetime.strftime('%Y-%m-%d %H:%M:%S')
        
    def get_weekday(self):
        """
        From 1 to 7
        """
        return self.datetime.isoweekday()
    
    def get_month(self):
        """
        From 1 to 12
        """
        return self.datetime.month
        
    def get_hours(self):
        return int(self.datetime.strftime('%H'))
    
    def get_minutes(self):
        return int(self.datetime.strftime('%M'))

    def __format__(self, *args):
        return self.str_datetime

    def __str__(self):
        return self.str_datetime


class str_time:
    def __init__(self, secs):
        self.str_time = _gmtime(int(secs))  # time.strftime('%H:%M:%S', time.gmtime(int(secs)))

    def __str__(self):
        return self.str_time

class str_resources:
    SEPARATOR = '#'
    REGEX = '[\d+;|' + SEPARATOR + ']+'
    REGEX_GROUP = '(?P<{}>[\d+;|' + SEPARATOR + ']+)'

    def __init__(self, nodes, resources):
        self.nodes = nodes
        self.resources = resources  # namedtuple('resources', [k for k in resources.keys()])(**resources)
        self.constants = CONSTANT()
        if not hasattr(self.constants, 'resource_order'):
            default_order = list(self.resources.keys())
            self.constants.load_constant('resource_order', default_order)
        self.order = self.constants.resource_order

    def __str__(self):
        return self.SEPARATOR.join(
            [';'.join([node.split('_')[1]] + [str(self.resources[_k]) for _k in self.order]) for node in
             self.nodes]) + self.SEPARATOR


CUSTOM_TYPES = {
    'accasim.utils.misc.str_datetime': str_datetime.REGEX_GROUP,
    'accasim.utils.misc.str_resources': str_resources.REGEX_GROUP,
}

class str_nodes:
    def __init__(self, nodes):
        self.nodes = nodes

    def __format__(self, format_spec):
        return self.__str__()

    def __str__(self):
        return ','.join([node.split('_')[1] for node in self.nodes])

class FrozenDict(Mapping):
    """

    Inmutable dictionary useful for storing parameters that are dinamycally loaded

    """

    def __init__(self, *args, **kwargs):
        self._d = dict(*args, **kwargs)
        self._hash = None

    def __iter__(self):
        return iter(self._d)

    def __len__(self):
        return len(self._d)

    def __getitem__(self, key):
        return self._d[key]

    def __hash__(self):
        if self._hash is None:
            self._hash = 0
            for pair in self.iteritems():
                self._hash ^= hash(pair)
        return self._hash
    
    def __str__(self):
        return str(self._d)
        
class SystemStatus:
    """
    
    Wathcer Daemon allows to track the simulation process through command line querying.
    
    """
    MAX_LENGTH = 2048

    def __init__(self, port, functions):
        """
    
        SystemStatus daemon constructor
        
        :param port: Port of the SystemStatus daemon
        :param functions: Available functions to call for data.
    
        """
        self.server_address = ['', port]
        af = socket.AF_INET
        self.sock = socket.socket(af, socket.SOCK_STREAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.settimeout(1)
        self.thread = None
        self.hastofinish = False
        self.const = CONSTANT()
        self.functions = functions

    def start(self):
        """
    
        Start the daemon
    
        """
        self.thread = _Thread(target=self.listenForRequests)
        self.hastofinish = False
        self.thread.start()

    def listenForRequests(self):
        """
    
        Listening for requests
    
        """
        # Listen for incoming connections
        # We retry binding the socket as long as we don't find a valid, unused port
        while True:
            try:
                self.sock.bind(tuple(self.server_address))
                break
            except socket.error:
                self.server_address[1] += 1

        self.sock.listen(5)

        while not self.hastofinish:
            try:
                connection, client_address = self.sock.accept()
                with connection:
                    # print('connection from %s' % (client_address[0]))
                    data = _loads(connection.recv(self.MAX_LENGTH).decode())
                    if isinstance(data, str):
                        response = {}
                        response['actual_time'] = str(str_datetime(self.call_inner_function('current_time_function')))
                        if data == 'progress':
                            response['input_filepath'] = self.const.WORKLOAD_FILEPATH
                            response['progress'] = _path.getsize(self.const.SCHEDULING_OUTPUT) / _path.getsize(
                                self.const.WORKLOAD_FILEPATH)
                            response['time'] = clock() - self.const.start_simulation_time
                        elif data == 'usage':
                            response['simulation_status'] = self.call_inner_function('simulated_status_function')
                            response['usage'] = self.call_inner_function('usage_function')
                        elif data == 'all':
                            response['input_filepath'] = self.const.input_filepath
                            response['progress'] = _path.getsize(self.const.sched_output_filepath) / _path.getsize(
                                self.const.input_filepath)
                            response['time'] = clock() - self.const.start_simulation_time
                            response['simulation_status'] = self.call_inner_function('simulated_status_function')
                            response['usage'] = self.call_inner_function('usage_function')
                        connection.sendall(_dumps(response).encode())
                    connection.close()
            except socket.timeout:
                pass
        self.sock.close()

    def call_inner_function(self, name):
        """
    
        Call a function and retrives it results
    
        :param name: name of the function 
    
        """
        if name in self.functions:
            _func = self.functions[name]
            if callable(_func):
                return _func()
            else:
                return _func
        raise Exception('{} was no defined'.format(name))

    def stop(self):
        """
    
        Stop the daemon
    
        """
        self.hastofinish = True
        if hasattr(self, 'timedemon'):
            self.timedemon.stop()



"""
MIT License

Copyright (c) 2017 cgalleguillosm

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.cm as mplcm
import matplotlib.colors as colors
import time
import matplotlib.animation as Animation
from accasim.utils.misc import str_datetime
from sortedcontainers import SortedListWithKey as sorted_object_list
from collections import namedtuple
from threading import Thread

class SystemUtilization:
    """
    Using the matplotlib animation feature, this class will create a plot and update it during the simulation execution.
    """
    
    def __init__(self, datasource, system_info, min_time=10 * 60, max_time=30 * 60):
        """
            :param datasource: The data to be plot. This must include the running jobs with the specific allocations. 
            :param system_info: The system resourcce to divide the plot area into resources types. Corresponds to the resource_manager system capacity.
            :param min_time: Default 600. How much the plot will show before the current time. In secs. 
            :param max_time: Default 1800. How much the plot will show after the current time. In secs.
        """
        self.datasource = datasource
        self.system_info = system_info
        self.min_time = min_time
        self.max_time = max_time
        self.job_patches = {}
        self.job_object = {}
        self.job_resouce_type = {}
        self.sorted_jobs = {}
        self.job_color = {}
        for key in self.system_info.keys():
            self.job_resouce_type[key] = namedtuple('Resource_{}'.format(key), ['id', key, 'expected_duration'])
            self.sorted_jobs[key] = sorted_object_list({'main': key, 'break_tie':'expected_duration'})
        self.figure = None
        self.animation = None
        self.axis_mapper = None
        self.color_pos = None
        self.last_line = {'time': None}
        self.color_blacklist = ['red']
        self.canvas = None
        
        # Create initial structure for ploting
        self.set_colors()
        self.running = False

    def start(self):
        """
        Initialize the plot
        """
        self.running = True
        self.create_empty_map()

    def stop(self):
        """
        Stops the animation and close the figure.
        """
        if self.running:
            self.running = False
            try:
                self.animation._stop()
                self.animation.repeat = False
            except AttributeError as e:
                print(e)
        plt.close()        
        
    def load_new_jobs(self, current_time, running_jobs, _resources):
        """
            Loads new jobs into the jobs that are "running" in the plot. Checks first if it is already in the plot, if not it is added.
            The jobs are sorted in regards of the resource utilization.
            :param current_time: Current simulation time.
            :param running_jobs: list of running jobs.
            :param _resources:  
        """
        for _id, _job in running_jobs.items():
            if _id in self.job_patches:
                continue
            self.job_object[_id] = _job
            for res in _resources:
                # self.sorted_jobs[res].add(self.job_resouce_type[res](_id, _job.get_total_requested_resource()[res], _job.expected_duration))
                self.sorted_jobs[res].add(self.job_resouce_type[res](_id, getattr(_job, res), _job.expected_duration))
        
    def job_draw(self, i):
        """
        Using the matplotlib animation the new jobs are draw into the plot. Besides previous jobs, axis info, legends are updated.
         
        """
        running_jobs = self.datasource['running_jobs']
        current_time = self.datasource['current_time']
        if current_time is None:
            self.stop()
            return
        self.xlimit(current_time)
        if not running_jobs:
            _finished = list(self.job_patches.keys())
            if _finished:
                self.update_finished_jobs(_finished)
            for res, ax in self.axis_mapper.items():
                self.modify_legend(ax, res) 
                self.update_ylabel(ax, res, 0)            
            
            return 
            
        _resources = self.system_info.keys()
        self.start_line_draw(current_time)
        
        # Load new jobs 
        self.load_new_jobs(current_time, running_jobs, _resources)
                
        # Remove jobs already finished
        _finished = [_id for _id in self.job_patches.keys() if _id not in running_jobs ]  # filter(lambda x: x not in running_jobs, list(self.job_patches.keys()))
        if _finished:
            self.update_finished_jobs(_finished)
            
        # Create new Rectangles or update its position (By increasing consumption)
        _stop_f = False
        for res in _resources:
            ax = self.axis_mapper[res]
            # Sorted by consumption
            _sorted_jobs = self.sorted_jobs[res].get_reversed_object_list()
            # Start from 0
            self.system_info[res]['used'] = 0
            
            for sjob in _sorted_jobs:
                _id = sjob.id
                job = self.job_object[_id]
                j_used = getattr(sjob, res)
                if j_used == 0:
                    # If 0 does not exists the patch
                    continue
                # Check if already exists
                if _id in self.job_patches:
                    _patches = self.job_patches[_id]
                    # Check if the patch exists
                    if res in _patches:
                        self.update_running_job(_patches, res, j_used)
                        continue
                else:
                    self.job_patches[_id] = {}
                    self.job_color[_id] = self.next_color()                
                
                a_used = self.system_info[res]['used']  # / self.system_info[res]['total']
                # _min, _max = self.job_normalization(current_time, sjob)
                r = patches.Rectangle(
                    (job.start_time, a_used / self.system_info[res]['total']), sjob.expected_duration, j_used / self.system_info[res]['total'],
                    fill=True, linewidth=0, edgecolor='black', facecolor=self.job_color[_id], linestyle='dashed', label=_id
                )                 
                self.job_patches[_id][res] = r 
                self.system_info[res]['used'] += j_used
                ax.add_patch(r)        
                 
            self.modify_legend(ax, res)
            self.update_ylabel(ax, res, self.system_info[res]['used'])
        plt.rcParams.update({'font.size': 16})
        self.figure.canvas.draw()
        self.figure.canvas.flush_events()             
        
    def update_ylabel(self, _ax, _res, _used):
        """
         Update the usage in the corresponding label of the resource type.
         
         :param _ax: Axis object
         :param _res: Axis resource name
         :param _used: Current usage of the resource type.  
        """
        _res_usage = _used / self.system_info[_res]['total']
        _ax.set_ylabel('{} {:.2%}'.format(_res, _res_usage), rotation=90, size=20)
        
    def modify_legend(self, ax, res, legend_size=10):
        """
         Modify the lenged of the resource type in regards of the jobs that use most the resources.
         Only a subset of jobs will be in the legend. By default 10.
         
         :param ax: Axis object
         :param res: Axis resource name
         :param legend_size: Default 10. Size of the legend.
        """
        h, l = ax.get_legend_handles_labels()
        # L.get_texts()[0].set_text('make it short')
        if len(l) <= legend_size:
            ax.legend(h, l, prop={'size':10}, loc='upper right', title='Job id')
            return
        _generated_legend = {l: h for h, l  in zip(h, l)}
        _modified_legend = []
        for _idx, _job in enumerate(self.sorted_jobs[res].get_reversed_object_list()):
            if _idx == legend_size:
                break
            _modified_legend.append((str(_job.id), _generated_legend[str(_job.id)]))
        l, h = zip(*_modified_legend)
        ax.legend(h, l, prop={'size':12}, loc='upper left')
        
    def update_running_job(self, _patches, _res, _used):
        """
        Updates the position of the patch in regards of the utilization of the resource
        
        :param _patches: Patch objects corresponding to a job.
        :param _res: Name of the resource type
        :param _used: How much the job uses of the resoure type. 
        """
        _patch = _patches[_res]
        _patch.set_y(self.system_info[_res]['used'] / self.system_info[_res]['total'])
        self.system_info[_res]['used'] += _used 
        
    def update_finished_jobs(self, _finished_jobs):
        """
         Patches of completed jobs are removed from the chart
         :param _finished_jobs: Array list of job's Id            
        """        
        for id in _finished_jobs:
            _dict_patches = self.job_patches.pop(id, {})
            
            for res, _list in self.sorted_jobs.items(): 
                _job = _list.pop(id=id)
                self.system_info[res]['used'] -= getattr(_job, res)
                if res in _dict_patches:
                    _dict_patches[res].remove()
    
            del self.job_color[id]
            del self.job_object[id]
        
    def job_normalization(self, current_time, job):
        """
        Jobs are normalized in terms of the current visualization, setting as min and max the maximum width of the patch
        :param current_time: Current time of the simulation.
        :param job: Job object.
        
        :return: Return the min and max positions of the patch  
        """
        _min = self.min_time
        _max = self.max_time
        _job_start = (current_time - job.start_time)
        _job_end = (job.start_time + job.expected_duration - current_time)
        if _job_start > _min:
             _min = _job_start
        if _job_end < _max:
            _max = _job_end
        return (_min, _max)
    
    def start_line_draw(self, current_time):
        """
        Draws the current time line and add the corresponding text
        :param current_time: Current simulation time 
        """
        if self.last_line['time'] == current_time:
            return
        self.last_line['time'] = current_time
        _text = self.last_line.pop('text', None)
        if _text is not None:
            _text.remove()
        self.last_line['text'] = self.figure.text(0.62, 0.05, 'Actual time: {}'.format(str_datetime(current_time)), fontsize=12, backgroundcolor='white')
        for name, ax in self.axis_mapper.items():
            _old = self.last_line.pop(name, None)
            if _old is not None:
                _old.remove()
            self.last_line[name] = ax.axvline(x=current_time, color='red')
        
            
    def create_empty_map(self):
        """
         Creates the empty matplotlib figure.
        """
        sys_config = self.system_info
        for _k in sys_config:
            sys_config[_k]['used'] = 0
        types = len(sys_config)
        self.figure, axis = plt.subplots(types, 1, sharex=True, sharey=True, figsize=(10, 8))
        sys_plot = {}
        for (ax, _sys) in zip(axis, sys_config):
            self.xlimit(0, ax)
            ax.set_ylabel(_sys, rotation=90, size=12)  # , size='large')
            ax.grid()
            xticks = ax.yaxis.get_major_ticks() 
            xticks[0].label1.set_visible(False)
            xticks[-1].label1.set_visible(False)
            sys_plot[_sys] = ax            

            for label in (ax.get_xticklabels() + ax.get_yticklabels()):
                label.set_fontsize(16)
        self.axis_mapper = sys_plot
        self.figure.subplots_adjust(hspace=0)
        plt.setp([a.get_xticklabels() for a in self.figure.axes[:-1]], visible=False)
        self.animation = Animation.FuncAnimation(self.figure, self.job_draw, interval=1000)
        
        plt.rcParams.update({'font.size': 16})
        plt.show()
                    
    def xlimit(self, current_time, ax=None):
        """
        Limit of the X Axis
        :param current_time: Current simulation time
        :param ax: None for all axis (there is one axis for each resource type) updates. Specify axis for a single update.   
        """
        if ax is None:         
            for _, ax in self.axis_mapper.items():
                self._set_xlimit(ax, current_time)
        else: 
            self._set_xlimit(ax, current_time)
            
    def _set_xlimit(self, ax, current_time):
        """
            Performs the xlimit update
            :param ax: Axis object to be updated
            :param current_time: Current simulation time  
        """
        if not current_time:
            return
        self.xlim = (current_time - self.min_time, current_time + self.max_time)
        ax.set_xlim(self.xlim)
        # -10min, Current, +10min, 20min 30min
        labels = ['-600', '-{}'.format(current_time % 600), '+600', '+1200', '+1800' ]
        ax.set_xticklabels(labels)
            
    
    def next_color(self):
        """
        Moves for all available matplotlib colors
        
        :return: Return an specific color to represent a job. 
        """
        if self.color_pos is None: 
            self.color_pos = len(self.colors) // 2
        while True:
            _next_pos = self.color_pos
            self.color_pos += 1
            if self.color_pos == len(self.colors):
                self.color_pos = 0 
            if self.colors[_next_pos] not in self.color_blacklist:
                break
        return self.colors[_next_pos] 
        
    def set_colors(self):
        """
        Sets the available colors 
        """
        self.colors = [name for name, hex in colors.cnames.items()]
